"""
Based on https://github.com/cctbx/cctbx_project, governed on the following license:

cctbx.license.txt.
"""

__all__ = [
    "RotamerScore",
    "RamaScore",
    "get_fraction_of_rotamer_outliers",
    "get_fraction_of_rama_outliers",
]

import functools
import math
import pickle
import tarfile
import warnings
import zipfile
from dataclasses import asdict, dataclass, field
from enum import IntEnum, StrEnum
from pathlib import Path
from typing import Any
import numpy as np
import requests
from biotite import structure as struc
from biotite.structure import check_res_id_continuity
from scipy.interpolate import RegularGridInterpolator

_ROTAMER_BASE_DIR = Path.home() / ".local/share/top8000_lib"
_ROTAMER_DIR = _ROTAMER_BASE_DIR / "reference_data-master" / "Top8000"
_ROTA_OUTLIER_THRESHOLD = 0.003
_ROTA_ALLOWED_THRESHOLD = 0.02
_RAMA_FAVORED_THRESHOLD = 0.02
_RAMA_ALLOWED_THRESHOLD = 0.001
_RAMA_GENERAL_ALLOWED_THRESHOLD = 0.0005
_RAMA_CISPRO_ALLOWED_THRESHOLD = 0.0020


class RamaResidueType(StrEnum):
    """
    Enum for residue types used in the Ramachandran library.
    This enum defines the types of residues that can be used with the Ramachandran phi/psi grid data.
    """

    GLY = "gly"
    CISPRO = "cispro"
    TRANSPRO = "transpro"
    PREPRO = "prepro"
    ILEVAL = "ileval"
    GENERAL = "general"


class RotamerGridResidueMap(StrEnum):
    """
    Enum for residue names used in the rotamer (chi angles) grid data.
    This enum maps residue names to their corresponding tags used in the rotamer library.
    """

    ALA = "NA"
    GLY = "NA"  # ALA and GLY have no chi angles
    ARG = "arg"
    ASN = "asn"
    ASP = "asp"
    CYS = "cys"
    GLN = "gln"
    GLU = "glu"
    HIS = "his"
    ILE = "ile"
    LEU = "leu"
    LYS = "lys"
    MET = "met"
    PHE = "phetyr"
    PRO = "pro"
    SER = "ser"
    THR = "thr"
    TRP = "trp"
    TYR = "phetyr"
    VAL = "val"


class ConformerClass(IntEnum):
    """
    Enum for conformer classification types.

    This enum defines the possible classifications for different conformations [1]_.
    It can be one of the following:
        - "FAVORED": The residue is in a favored rotamer conformation.
        - "ALLOWED": The residue is in an allowed rotamer conformation.
        - "OUTLIER": The residue is in an outlier rotamer conformation.
        - "UNKNOWN": The residue's rotamer score could not be determined.

    References
    ----------
    .. [1] https://doi.org/10.1002/prot.25039
    """

    FAVORED = 1
    ALLOWED = 2
    OUTLIER = 3
    UNKNOWN = 4


@dataclass
class DihedralScore:
    """
    A dataclass to hold the results of rotamer and rama score checks.

    Attributes
    ----------
    pct : float
        The percentile score of the rotamer or rama check.
    classification : ConformerClass
        The classification of the score (e.g., FAVORED, ALLOWED, OUTLIER).

    References
    ----------
    .. [1] https://pmc.ncbi.nlm.nih.gov/articles/PMC4983197/
    """

    pct: float
    classification: ConformerClass


def _get_mmtbx_neg180_to_180_value(angle_deg: float) -> float:
    """
    Convert phi and psi angles to the range [-180, 180).

    Parameters
    ----------
    angle_deg : float
        The angle in degrees.

    Returns
    -------
    float
        The converted angle in the range [-180, 180).

    Notes
    -----
    Taken from https://github.com/cctbx/cctbx_project/blob/ee756cb47e375e11b4c37b65c418747f42104446/mmtbx/geometry_restraints/ramachandran.h#L230C7-L244C1
    """
    if angle_deg > 180:
        angle_deg -= 360
    elif angle_deg <= -180:
        angle_deg += 360
    assert -180 <= angle_deg <= 180, (
        f"Angle must be in the range [-180, 180), but we got {angle_deg}"
    )
    return angle_deg


def _wrap_phi_psi(phi: float, psi: float) -> tuple[float, float]:
    """
    Wrap phi and psi angles to the range [-180, 180).

    Parameters
    ----------
    phi : float
        The phi angle in degrees.
    psi : float
        The psi angle in degrees.

    Returns
    -------
    tuple
        A tuple containing the converted phi and psi angles.

    Notes
    -----
    Taken from https://github.com/cctbx/cctbx_project/blob/ee756cb47e375e11b4c37b65c418747f42104446/mmtbx/geometry_restraints/ramachandran.h#L230C7-L244C1
    """
    return _get_mmtbx_neg180_to_180_value(phi), _get_mmtbx_neg180_to_180_value(psi)


def _wrap_symmetrical(
    residue_tag: RotamerGridResidueMap, wrapped_chis: list[float]
) -> list[float]:
    """
    Apply symmetrical wrapping to chi angles for specific amino acids.

    Parameters
    ----------
    residue_tag : RotamerGridResidueMap
        The residue tag enum
    wrapped_chis : list[float]
        The list of wrapped chi angles.

    Returns
    -------
    list[float]
        The list of symmetrically wrapped chi angles.

    Notes
    -----
    Taken from Source: https://github.com/cctbx/cctbx_project/blob/ee756cb47e375e11b4c37b65c418747f42104446/mmtbx/rotamer/rotamer_eval.py#L368
    """
    residue_tag = residue_tag.lower()
    if residue_tag == "asp" or residue_tag == "glu" or residue_tag == "phetyr":
        i = len(wrapped_chis) - 1
        if wrapped_chis[i] is not None:
            wrapped_chis[i] = wrapped_chis[i] % 180
            if wrapped_chis[i] < 0:
                wrapped_chis[i] += 180
    return wrapped_chis


def _wrap_chis(
    residue_tag: RotamerGridResidueMap, chis: list[float], symmetry: bool = True
) -> list[float]:
    """
    Wrap chi angles to the range [0, 360) or [-180, 180) depending on the residue type.

    Parameters
    ----------
    residue_tag : RotamerGridResidueMap
        The residue tag enum
    chis : list[float]
        The list of chi angles.
    symmetry : bool
        Whether to apply symmetrical wrapping.

    Returns
    -------
    list[float]
        The list of wrapped chi angles.

    Notes
    -----
    Taken from Source: https://github.com/cctbx/cctbx_project/blob/ee756cb47e375e11b4c37b65c418747f42104446/mmtbx/rotamer/rotamer_eval.py#L368
    """
    residue_tag = residue_tag.lower()
    wrapped_chis = []
    for i in range(0, len(chis)):
        if chis[i] is not None:
            wrapped_chis.append(chis[i] % 360)
            if wrapped_chis[i] < 0:
                wrapped_chis[i] += 360
        else:
            wrapped_chis.append(None)
    if symmetry:
        wrapped_chis = _wrap_symmetrical(residue_tag, wrapped_chis)
    return wrapped_chis


def _download_and_extract(url: str, dest_path: Path | None = None) -> Path:
    """
    Download a file from the given URL and extract it if it's a zip or tar archive.
    If `dest_path` is provided, it will be used as the destination file path.
    If `dest_path` is None, the file will be saved in the current directory with the same name as the URL.

    Parameters
    ----------
    url : str
        The URL to download the file from.
    dest_path : Path
        The destination path where the file should be saved. If None, uses the URL's name.

    Returns
    -------
    Path
        The path where the file was saved or extracted.
    """
    if dest_path is None:
        dest_path = Path(Path(url).name)
    else:
        dest_path = Path(dest_path)
    # Ensure parent directory exists
    dest_path.parent.mkdir(parents=True, exist_ok=True)
    # Download the file
    if dest_path.exists():
        return dest_path.parent
    r = requests.get(url)
    r.raise_for_status()
    data = r.content
    if dest_path:
        with open(dest_path, "wb") as f:
            f.write(data)
        # Check if it's a zip or tar file
        if str(dest_path).endswith(".zip"):
            with zipfile.ZipFile(dest_path, "r") as zip_ref:
                zip_ref.extractall(dest_path.parent)
            return dest_path.parent
        elif str(dest_path).endswith(".tar.gz") or str(dest_path).endswith(".tgz"):
            with tarfile.open(dest_path, "r:gz") as tar_ref:
                tar_ref.extractall(dest_path.parent)
            return dest_path.parent
    return dest_path.parent


def _get_residue_chis(
    atom_array: struc.AtomArray, res_id_mask: np.ndarray
) -> list[float]:
    """
    Compute the chi angles for a given residue.

    Parameters
    ----------
    atom_array : AtomArray
        The structure
    res_id_mask : np.ndarray
        A boolean mask selecting atoms of the residue (or an index for residue).

    Returns
    -------
    tuple[float]
        A tuple of chi angles (chi1, chi2, chi3, chi4). Missing chis will be omitted.

    Notes
    -----
    The chi angles are computed based on the standard definitions for each amino acid.
    In case of missing atoms, the corresponding chi angle will be omitted.
    """
    # We'll compute common chis by looking up named atoms for each residue type.
    res_atoms = atom_array[res_id_mask]
    # Build a map name -> coordinates
    coords = {
        name.strip(): coord for name, coord in zip(res_atoms.atom_name, res_atoms.coord)
    }
    # mapping of chi definitions per residue: set of atom name tuples (A,B,C,D) for chi1..chi4
    CHI_DEFS = {
        # Example: for ARG: chi1: N-CA-CB-CG  (names: 'N','CA','CB','CG' )
        "ARG": [
            ("N", "CA", "CB", "CG"),
            ("CA", "CB", "CG", "CD"),
            ("CB", "CG", "CD", "NE"),
            ("CG", "CD", "NE", "CZ"),
        ],
        "LEU": [
            ("N", "CA", "CB", "CG"),
            ("CA", "CB", "CG", "CD1"),
        ],  # LEU uses chi2 to CD1 or CD2 depending; this is simplified
        "VAL": [("N", "CA", "CB", "CG1")],
        "ILE": [("N", "CA", "CB", "CG1"), ("CA", "CB", "CG1", "CD1")],
        "MET": [
            ("N", "CA", "CB", "CG"),
            ("CA", "CB", "CG", "SD"),
            ("CB", "CG", "SD", "CE"),
        ],
        "LYS": [
            ("N", "CA", "CB", "CG"),
            ("CA", "CB", "CG", "CD"),
            ("CB", "CG", "CD", "CE"),
            ("CG", "CD", "CE", "NZ"),
        ],
        "PHE": [
            ("N", "CA", "CB", "CG"),
            ("CA", "CB", "CG", "CD1"),
        ],
        "TRP": [
            ("N", "CA", "CB", "CG"),
            ("CA", "CB", "CG", "CD1"),
        ],
        "TYR": [
            ("N", "CA", "CB", "CG"),
            ("CA", "CB", "CG", "CD1"),
        ],
        "ASN": [("N", "CA", "CB", "CG"), ("CA", "CB", "CG", "OD1")],
        "GLN": [
            ("N", "CA", "CB", "CG"),
            ("CA", "CB", "CG", "CD"),
            ("CB", "CG", "CD", "OE1"),
        ],
        "ASP": [("N", "CA", "CB", "CG"), ("CA", "CB", "CG", "OD1")],
        "GLU": [
            ("N", "CA", "CB", "CG"),
            ("CA", "CB", "CG", "CD"),
            ("CB", "CG", "CD", "OE1"),
        ],
        "CYS": [("N", "CA", "CB", "SG")],
        "HIS": [("N", "CA", "CB", "CG"), ("CA", "CB", "CG", "ND1")],
        "PRO": [("N", "CA", "CB", "CG"), ("CA", "CB", "CG", "CD")],
        "SER": [("N", "CA", "CB", "OG")],
        "THR": [("N", "CA", "CB", "OG1")],
        "ALA": [],  # ALA has no chi angles
        "GLY": [],  # GLY has no chi angles
    }
    # detect residue type name
    resname = res_atoms.res_name[0].upper()
    chis: list[float] = []
    if resname not in CHI_DEFS:
        return chis

    for i, atom_tuple in enumerate(CHI_DEFS[resname], start=1):
        try:
            p = [coords[a] for a in atom_tuple]
        except KeyError as e:
            warnings.warn(
                f"Skipping chi angle {i} for residue {resname}: {e} due to missing atoms"
            )
            # Skip if any atom is missing
            continue
        else:
            ang = math.degrees(struc.dihedral(*p))
            chis.append(ang)

    return chis


@functools.cache
def _download_top1000_lib() -> Path:
    """
    Download the Top8000 rotamer library from the Richardson Lab GitHub.
    This is a convenience function to download the entire library.

    Returns
    -------
    Path
        The path to the downloaded Top8000 rotamer library directory.
        If the directory already exists, it returns the existing path.
    """
    URL = "https://github.com/rlabduke/reference_data/archive/refs/heads/master.zip"
    if _ROTAMER_DIR.exists():
        return _ROTAMER_DIR
    data_dir = _download_and_extract(URL, _ROTAMER_BASE_DIR / "top1000.zip")
    if data_dir is None:
        raise RuntimeError("Failed to download or extract Top8000 rotamer library")
    return _ROTAMER_DIR


@functools.cache
def _load_contour_grid_text(resname_grid_data: Path | Any) -> dict[str, Any]:
    """
    Parse a Top8000 pct_contour_grid text file into a numpy grid and axis coordinate arrays.

    Parameters
    ----------
    resname_grid_data : Path | str
        The path to the grid data file or the raw text data.

    Returns
    -------
    dict[str, Any]
        A dictionary containing the grid data, axes, and wrap information.
        ```{
        'grid': np.ndarray,
        'axes': [axis1_centers, axis2_centers, ...],
        'wrap': [True/False per axis]
        }```.

    Notes
    -----
    The grid data file should have a header with the number of dimensions and axis information,
    followed by the grid values. The format is expected to be:
    # number of dimensions: 3
    # x1: 0.0  360.0  36 true
    # x2: -180.0  180.0  36 true
    # x3: 0.0  360.0  36 false
    0.0 0.0 0.0 0.01
    The first line contains the number of dimensions, and each subsequent line describes an axis
    with its low and high bounds, number of bins, and whether it wraps around (true/false).
    If the file is not found or cannot be parsed, it returns None.
    If the file is a string, it will be treated as raw text data.
    If the file is a Path, it will be read from disk.
    If the file is not found or cannot be parsed, it returns None.
    """
    # Download data
    if isinstance(resname_grid_data, Path):
        with open(resname_grid_data, "r") as f:
            text = f.read()
    else:
        text = resname_grid_data
    lines = [ln.strip() for ln in text.splitlines() if ln.strip()]
    # Parse header
    axes_meta = []
    data_start_idx = 0
    for i, ln in enumerate(lines):
        if ln.lower().startswith("# number of dimensions"):
            continue
        elif ln.strip().startswith("#   x"):
            parts = ln.split(":")[1].split()
            low, high, nbins, wrap = (
                float(parts[0]),
                float(parts[1]),
                int(parts[2]),
                parts[3].lower() == "true",
            )
            axes_meta.append((low, high, nbins, wrap))
        elif not ln.startswith("#"):
            data_start_idx = i
            break
    # Build coordinate arrays (bin centers)
    axes_coords = []
    wraps = []
    steps = []
    for low, high, nbins, wrap in axes_meta:
        step = (high - low) / nbins
        centers = np.linspace(low + step / 2, high - step / 2, nbins)
        axes_coords.append(centers)
        wraps.append(wrap)
        steps.append(step)

    # Prepare empty grid
    shape = [meta[2] for meta in axes_meta]
    grid = np.zeros(shape, dtype=float)

    # Fill grid from data lines
    for ln in lines[data_start_idx:]:
        parts = ln.split()
        coords = list(map(float, parts[:-1]))
        val = float(parts[-1])
        # Map coords to bin indices
        idxs = []
        for dim, (low, high, nbins, wrap) in enumerate(axes_meta):
            step = (high - low) / nbins
            bin_idx = (
                int((coords[dim] - low) / step) % nbins
                if wrap
                else int((coords[dim] - low) / step)
            )
            idxs.append(bin_idx)
        grid[tuple(idxs)] = val

    return {
        "grid": grid,
        "axes": axes_coords,
        "wrap": wraps,
        "span": np.array([am[:2] for am in axes_meta]),
        "steps": steps,
    }


@functools.cache
def _load_contour_grid_for_residue(
    grid_dirname: str,
    resname_tag: str,
    grid_data_tag: str = "rota8000",
    files_to_skip: tuple[str, ...] | None = None,
) -> dict[str, np.ndarray]:
    """
    Load contour grid for a given residue name from the Top8000 rotamer library.

    Parameters
    ----------
    grid_dirname : str
        The directory name where the contour grids are stored.
    resname_tag : str
        The residue name tag to load the contour grid for (e.g. "arg", "leu").
    grid_data_tag : str
        The grid data tag to load (default is "rota8000").
    files_to_skip : tuple[str, ...] | None
        A tuple of file names to skip when loading the grid data.

    Returns
    -------
    dict
        A dictionary containing the grid data, axes, and wrap information.
    """
    all_dict_contour = _load_all_contour_grid_from_pickle(
        grid_dirname=grid_dirname,
        grid_data_tag=grid_data_tag,
        files_to_skip=files_to_skip,
    )

    return all_dict_contour.get(resname_tag, {})


@functools.cache
def _generate_contour_grids_data(
    grid_dirname: str, grid_data_tag: str, files_to_skip: tuple[str, ...] | None = None
) -> None:
    """
    Generate contour grids data for the Top8000 rotamer library and save it to a pickle file.

    Parameters
    ----------
    grid_dirname : str
        The directory name where the contour grids will be saved.
    grid_data_tag : str
        The grid data tag to use (default is "rota8000").
    files_to_skip : tuple[str, ...] | None
        A tuple of file names to skip when generating the contour grids. If None, no files are skipped.

    Returns
    -------
    None
        This function saves the contour grids data to a pickle file.
        If the file already exists, it skips the generation.
    """
    output_path = _ROTAMER_DIR / f"{grid_dirname}.pkl"
    if output_path.exists():
        warnings.warn(
            f"Contour grids already generated at {output_path}. Skipping generation."
        )
        return
    data_dict = {}
    data_dir = _download_top1000_lib()
    if data_dir is None or not data_dir.exists():
        raise RuntimeError(
            "Top8000 rotamer library not found; cannot generate contour grids."
        )
    all_grid_files = list(data_dir.glob(f"{grid_dirname}/{grid_data_tag}-*.data"))

    if files_to_skip is not None:
        all_grid_files = [f for f in all_grid_files if f.name not in files_to_skip]

    if not all_grid_files:
        raise RuntimeError(f"No contour grid files found in {data_dir / grid_dirname}")
    for grid_file in all_grid_files:
        resname_tag = grid_file.stem.split("-")[
            1
        ].lower()  # e.g. "rota8000-arg.data" -> "arg"

        tmp_contour = _load_contour_grid_text(grid_file)
        if tmp_contour is None:
            continue
        data_dict[resname_tag] = tmp_contour
    with open(output_path, "wb") as f:
        pickle.dump(data_dict, f)


@functools.cache
def _load_all_contour_grid_from_pickle(
    grid_dirname: str, grid_data_tag: str, files_to_skip: tuple[str, ...] | None = None
) -> dict[str, dict[str, np.ndarray]]:
    """
    Load all contour grids from a pickle file for the given grid directory and data tag.
    If the pickle file does not exist, it generates the contour grids data and saves it to a pickle file.

    Parameters
    ----------
    grid_dirname : str
        The directory name where the contour grids are stored.
    grid_data_tag : str
        The grid data tag to load (default is "rota8000").
    files_to_skip : tuple[str, ...] | None
        A tuple of file names to skip when loading the grid data. If None, no files are skipped.

    Returns
    -------
    dict[str, dict[str, np.ndarray]]
        A dictionary containing the loaded contour grids.
    """
    contour_path = _ROTAMER_DIR / f"{grid_dirname}.pkl"
    if not contour_path.exists():
        _generate_contour_grids_data(
            grid_dirname=grid_dirname,
            grid_data_tag=grid_data_tag,
            files_to_skip=files_to_skip,
        )
    with open(contour_path, "rb") as f:
        data = pickle.load(f)
    return data


def _truncate_angles_within_a_step_of_grid_span(
    coord: float, low: float, high: float, step: float
) -> float:
    """
    Truncate an angle to be within a step of the grid span.

    This function ensures that the coordinate is within the specified low and high bounds,
    and if it is outside, it will return the closest bound if it is within one step of it.

    Parameters
    ----------
    coord : float
        The coordinate to truncate.
    low : float
        The lower bound of the grid span.
    high : float
        The upper bound of the grid span.
    step : float
        The step size of the grid.

    Returns
    -------
    float
        The truncated coordinate, which is either within the bounds or the closest bound if within one step.
    """
    # Handle edge cases
    if coord < low:
        if abs(coord - low) < step:
            return low
    elif coord > high:
        if abs(coord - high) < step:
            return high
    return coord


def _identify_gaps_in_atom_array(
    atom_array: struc.AtomArray,
) -> np.ndarray:
    """
    Identify gaps in the atom array based on missing atoms or residues.

    Parameters
    ----------
    atom_array : AtomArray
        The atom array to check for gaps.

    Returns
    -------
    np.ndarray
        An array of residue IDs where gaps are identified.
    """
    gap_ending_atom_idx = check_res_id_continuity(atom_array)
    # Add the preceeding atoms
    gap_both_atom_idx = np.concatenate([gap_ending_atom_idx, gap_ending_atom_idx - 1])
    return atom_array.res_id[gap_both_atom_idx]


def _interp_wrapped(
    resname_tag: str,
    grid_obj: dict[str, np.ndarray],
    coords_deg: list[float],
    coords_type: str,
) -> tuple[float, list[float]]:
    """
    Interpolate probability at given χ coords (deg), handling wrapping axes.

    Parameters
    ----------
    grid_obj : dict[str, np.ndarray]
        The grid object containing 'grid', 'axes', and 'wrap' information.
    coords_deg : list[float]
        List of coordinates in degrees, which could be [phi, psi] / [χ1, χ2, ...].
    coords_type : str
        The type of coordinates being interpolated (e.g., "phi-psi" or "chi").

    Returns
    -------
    tuple[float, list[float]]
        A tuple containing:
        - The interpolated value at the given coordinates.
        - The wrapped coordinates used for interpolation.
    """
    coords = []
    if coords_type not in ["phi-psi", "chi"]:
        raise ValueError(
            f"Invalid coords_type: {coords_type}. Expected 'phi-psi' or 'chi'."
        )
    if np.isnan(coords_deg).any():
        return np.nan, coords_deg

    if (coords_type == "phi-psi") and all(grid_obj["wrap"]):
        # Wrap phi and psi angles to [-180, 180)
        coords = list(_wrap_phi_psi(coords_deg[0], coords_deg[1]))

    elif (coords_type == "chi") and all(grid_obj["wrap"]):
        # Wrap chi angles to [0, 360) or [0, 180) depending on residue type
        coords = list(_wrap_chis(resname_tag, coords_deg, symmetry=True))

    coords = [
        _truncate_angles_within_a_step_of_grid_span(
            coord, grid_obj["axes"][i][0], grid_obj["axes"][i][-1], grid_obj["steps"][i]
        )
        for i, coord in enumerate(coords)
    ]

    interp_func = RegularGridInterpolator(
        tuple(grid_obj["axes"]), grid_obj["grid"], bounds_error=False, fill_value=np.nan
    )
    if interp_func(coords)[0] is np.nan:
        warnings.warn(
            f"Interpolated value is NaN for residue {resname_tag} at coords {coords}. Check if coords are within grid span {grid_obj['span']}."
        )

    return float(interp_func(coords)[0]), coords


def _check_rotamer(
    atom_array: struc.AtomArray, res_id: int, chain_id: str
) -> tuple[DihedralScore, list[float]]:
    """
    Check the rotamer classification for a given residue in the atom array.

    Parameters
    ----------
    atom_array : struc.AtomArray
        The structure
    res_id : int
        The residue ID to check.
    chain_id : str
        The chain ID to check.

    Returns
    -------
    tuple[DihedralScore, list[float]]
        A tuple containing:
        - DihedralScore: The rotamer score and classification.
        - dict[str, float]: The observed chi angles for the residue.
    """
    mask = (atom_array.chain_id == chain_id) & (atom_array.res_id == res_id)
    if not np.any(mask):
        raise ValueError("Residue not found")

    resname = atom_array.res_name[mask][0].upper()
    observed = _get_residue_chis(atom_array, mask)

    if not observed:
        return DihedralScore(
            pct=np.nan, classification=ConformerClass.UNKNOWN
        ), observed

    resname_tag = RotamerGridResidueMap.__dict__.get(resname, None)

    if resname_tag is None or resname_tag == "NA":
        warnings.warn(f"Rotamer classification not available for residue {resname}.")
        return DihedralScore(
            pct=np.nan, classification=ConformerClass.UNKNOWN
        ), observed

    grid_obj = _load_contour_grid_for_residue(
        grid_dirname="Top8000_rotamer_pct_contour_grids",
        resname_tag=resname_tag,
        grid_data_tag="rota8000",
        files_to_skip=("rota8000-leu-raw.data",),  # Use rota8000-leu.data instead
    )

    # Build coords in same order as grid dims
    coords_list = []
    for i in range(len(grid_obj["axes"])):
        coords_list.append(observed[i] if i < len(observed) else 0.0)

    pct, wrapped_angles = _interp_wrapped(resname_tag, grid_obj, coords_list, "chi")

    if np.isnan(pct):
        classification = (
            ConformerClass.FAVORED
        )  # In mmtbx.validation.rotalyze, they counted this as FAVORED

    elif pct >= _ROTA_ALLOWED_THRESHOLD:
        classification = ConformerClass.FAVORED
    elif pct >= _ROTA_OUTLIER_THRESHOLD:
        classification = ConformerClass.ALLOWED
    else:
        classification = ConformerClass.OUTLIER

    return DihedralScore(pct=pct, classification=classification), wrapped_angles


@dataclass
class ResidueRotamerScore:
    """
    Rotamer score for a residue in a protein structure.

    This class represents the rotamer score for a single residue, including its name,
    ID, observed chi angles, rotamer score percentage, and classification.

    Attributes
    ----------
    resname : str
        The name of the residue (e.g. "ARG", "LEU").
    resid : int
        The ID of the residue in the structure.
    chain_id : str
        The chain ID of the residue.
    angles : dict[str, float]
        A dictionary containing the observed chi angles for the residue, e.g. {'chi1': 60.0, 'chi2': -45.0}.
    pct : float
        The rotamer score percentage for the residue, e.g. 0.01 for 1% favored.
    classification : ConformerClass
        The classification of the residue based on its rotamer score percentage.

    References
    ----------
    .. [1] https://pmc.ncbi.nlm.nih.gov/articles/PMC4983197/
    """

    res_name: str
    res_id: int
    chain_id: str
    angles: dict[str, float]
    pct: float
    classification: ConformerClass

    @staticmethod
    def from_residue(
        atom_array: struc.AtomArray, res_id: int, res_name: str, chain_id: str
    ) -> "ResidueRotamerScore":
        """
        Create ResidueRotamerScore from a residue object.

        Parameters
        ----------
        atom_array : struc.AtomArray
            The structure.
        res_id : int
            The residue ID to check.
        res_name : str
            The residue name to check (e.g. "ARG", "LEU").
        chain_id : str
            The chain ID to check.

        Returns
        -------
        ResidueRotamerScore
            An instance of ResidueRotamerScore containing the rotamer score for the residue.
        """
        score, angles = _check_rotamer(atom_array, res_id, chain_id)
        chis = {f"chi{i + 1}": ang for i, ang in enumerate(angles)}

        return ResidueRotamerScore(
            res_name=res_name,
            res_id=res_id,
            chain_id=chain_id,
            angles=chis,
            **asdict(score),
        )


def _classify_rotamers(atom_array: struc.AtomArray) -> list["ResidueRotamerScore"]:
    """
    Create ResidueRotamerScore from a protein structure.

    Parameters
    ----------
    atom_array : struc.AtomArray
        The structure.

    Returns
    -------
    list[ResidueRotamerScore]
        An instance of RotamerScore containing the rotamer scores for the structure.
    """
    if all(atom_array.hetero):
        warnings.warn(
            "RotamerScore: Cannot be computed; atom_array must contain some protein residues"
        )
        return []
    atom_array = atom_array[
        struc.filter_canonical_amino_acids(atom_array) & ~atom_array.hetero
    ]
    if atom_array.array_length == 0:
        warnings.warn(
            "RotamerScore: Cannot be computed; atom_array must contain at least one residue"
        )
        return []
    rotamer_scores: list[ResidueRotamerScore] = []
    for chain_arr in struc.chain_iter(atom_array):
        chain_arr = chain_arr[struc.filter_canonical_amino_acids(chain_arr)]
        res_ids, res_names = struc.get_residues(chain_arr)
        rotamer_scores = []
        for res_id, res_name in zip(res_ids, res_names):
            chain_id = chain_arr.chain_id[chain_arr.res_id == res_id][0]

            if res_name not in RotamerGridResidueMap.__members__:
                continue
            rotamer_scores.append(
                ResidueRotamerScore.from_residue(
                    atom_array=atom_array,
                    res_id=res_id,
                    res_name=res_name,
                    chain_id=chain_id,
                )
            )

    return rotamer_scores


@dataclass
class RotamerScore:
    """
    Rotamer score for a given protein structure.
    This class represents the rotamer scores for all residues in a protein structure,
    including their names, IDs, observed chi angles, rotamer score percentages, and
    classifications.

    It is used to assess the quality of the protein structure based on the rotamer
    conformations and to identify potential outliers in the rotamer angles.

    Attributes
    ----------
    rotamer_scores : list[ResidueRotamerScore]
        A list of ResidueRotamerScore objects, each representing the rotamer score for a
        residue in the structure.
        Each ResidueRotamerScore contains the residue name, residue ID, chain ID,
        observed chi angles, rotamer score percentage and classification.

    References
    ----------
    .. [1] http://molprobity.biochem.duke.edu/help/validation_options/summary_table_guide.html
    .. [2] https://pmc.ncbi.nlm.nih.gov/articles/PMC4983197/
    """

    rotamer_scores: list[ResidueRotamerScore] = field(
        default_factory=list,
        metadata={
            "description": "List of rotamer scores for each residue in the structure."
        },
    )

    @staticmethod
    def from_atoms(atom_array: struc.AtomArray) -> "RotamerScore":
        """
        Create RotamerScore from a protein structure or a stack of structures.

        Parameters
        ----------
        atom_array : struc.AtomArray | struc.AtomArrayStack
            The AtomArray or AtomArrayStack containing the structure(s).

        Returns
        -------
        RotamerScore
            An instance of RotamerScore containing the rotamer scores for the structure(s).
        """
        return RotamerScore(rotamer_scores=_classify_rotamers(atom_array))


def _get_residue_phi_psi_omega(
    atom_array: struc.AtomArray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Get the phi, psi, and omega dihedral angles for a given residue.

    Parameters
    ----------
    atom_array : struc.AtomArray
        The structure.

    Returns
    -------
    dict[str, np.ndarray]
        A dictionary containing the phi, psi, and omega angles in degrees.
    """
    if atom_array.shape[0] < 4:
        raise ValueError(
            "atom_array must contain at least 4 atoms for dihedral angle calculation"
        )
    phi, psi, omega = struc.dihedral_backbone(atom_array)
    # Remove invalid values (NaN) at first and last position
    phi = phi[1:-1]
    psi = psi[1:-1]
    omega = omega[1:-1]

    # Convert to degrees, keeping NaNs as NaNs
    phi = np.where(np.isnan(phi), np.nan, np.rad2deg(phi))
    psi = np.where(np.isnan(psi), np.nan, np.rad2deg(psi))
    omega = np.where(np.isnan(omega), np.nan, np.rad2deg(omega))
    return phi, psi, omega


def _is_cislike_peptide(omega: float) -> bool:
    """
    Check if a peptide bond is cis-like based on the omega dihedral angle.
    A peptide bond is considered cis-like if the omega angle is between -30 and 30 degrees.
    This is based on the definition used in the mmtbx.validation.omegalyze module [1]_.

    Parameters
    ----------
    omega : float
        The omega dihedral angle in degrees.

    Returns
    -------
    bool
        True if the peptide bond is cis-like (omega between -90 and 90 degrees), False otherwise.

    References
    ----------
    .. [1] hhttps://github.com/cctbx/cctbx_project/blob/3de736be359daeba2a2f3d7da6a4a94faf4ec4d5/mmtbx/validation/omegalyze.py#L560C1-L564C20
    """
    if np.isnan(omega):
        return False
    if (omega > -30) and (omega < 30):
        return True
    return False


def _assign_rama_types(
    atom_array: struc.AtomArray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, list[str]]:
    """
    Assign Ramachandran types to residues based on their phi, psi, and omega angles.

    Parameters
    ----------
    atom_array : struc.AtomArray
        The structure.

    Returns
    -------
    dict
        A dictionary containing the phi, psi, omega angles and their corresponding
        Ramachandran types.

    Notes
    -----
    This function assumes `atom_array` is already cleaned and contains only relevant
    residues.
    It should not contain water or other non-protein residues.
    """
    # TODO: Make this function more performant by avoiding multiple calls to struc.dihedral_backbone
    phis, psis, omegas = _get_residue_phi_psi_omega(atom_array)

    _, res_names = struc.get_residues(atom_array)
    next_res_names = np.roll(res_names, -1)[
        1:-1
    ]  # Shifted by -1, remove first and last

    def rama_type(res_name: str, next_res_name: str, omega_val: float) -> str:
        if res_name == "GLY":
            return RamaResidueType.GLY
        elif res_name == "PRO":
            return (
                RamaResidueType.CISPRO
                if _is_cislike_peptide(omega_val)
                else RamaResidueType.TRANSPRO
            )
        elif next_res_name == "PRO":
            return RamaResidueType.PREPRO
        elif res_name in ["ILE", "VAL"]:
            return RamaResidueType.ILEVAL
        else:
            return RamaResidueType.GENERAL

    rama_types = [
        rama_type(res_name, next_res_name, omega_val)
        for res_name, next_res_name, omega_val in zip(
            res_names[1:-1], next_res_names, omegas
        )
    ]

    return phis, psis, omegas, rama_types


def _check_rama(
    phi: float,
    psi: float,
    resname_tag: RamaResidueType,
) -> "DihedralScore":
    """
    Check the Ramachandran classification for a given residue based on its phi, psi, and omega angles.

    Parameters
    ----------
    phi : float
        The phi angle in degrees.
    psi : float
        The psi angle in degrees.
    resname_tag : RamaResidueType
        The residue name tag enum

    Returns
    -------
    DihedralScore
        An instance of DihedralScore containing the Ramachandran score and classification.

    Notes
    -----
    This function uses the Top8000 Ramachandran contour grids to classify the residue.
    It checks the phi, psi, and omega angles against the grids and assigns a
    classification of "FAVORED", "ALLOWED", or "OUTLIER" based on the percentage of
    favored conformations.
    If the residue is CIS or PRO, it uses special handling for those residues.
    The thresholds for classification are defined as follows:
    - FAVORED: pct >= `_RAMA_FAVORED_THRESHOLD`
    - ALLOWED: pct >= `_RAMA_ALLOWED_THRESHOLD` (or `_RAMA_GENERAL_ALLOWED_THRESHOLD`
      for "general" residues, or `_RAMA_CISPRO_ALLOWED_THRESHOLD` for "cispro" residues)
    - OUTLIER: pct < `_RAMA_ALLOWED_THRESHOLD` (or `_RAMA_GENERAL_ALLOWED_THRESHOLD` for
      "general" residues, or `_RAMA_CISPRO_ALLOWED_THRESHOLD` for "cispro" residues)

    References
    ----------
    .. [1] https://pmc.ncbi.nlm.nih.gov/articles/PMC4983197/
    """

    grid_obj = _load_contour_grid_for_residue(
        grid_dirname="Top8000_ramachandran_pct_contour_grids",
        resname_tag=resname_tag,
        grid_data_tag="rama8000",
        files_to_skip=None,
    )
    # Build coords in same order as grid dims
    coords_list = [phi, psi]
    pct, _ = _interp_wrapped(resname_tag, grid_obj, coords_list, "phi-psi")
    if resname_tag == "general":
        if pct >= _RAMA_GENERAL_ALLOWED_THRESHOLD:
            classification = ConformerClass.ALLOWED
        else:
            classification = ConformerClass.OUTLIER
    elif resname_tag == "cispro":
        if pct >= _RAMA_CISPRO_ALLOWED_THRESHOLD:
            classification = ConformerClass.ALLOWED
        else:
            classification = ConformerClass.OUTLIER
    else:
        if pct >= _RAMA_ALLOWED_THRESHOLD:
            classification = ConformerClass.ALLOWED
        else:
            classification = ConformerClass.OUTLIER
    if pct >= _RAMA_FAVORED_THRESHOLD:
        classification = ConformerClass.FAVORED

    return DihedralScore(pct=pct, classification=classification)


@dataclass
class ResidueRamaScore:
    """
    Ramachandran score for a residue in a protein structure.
    This class represents the Ramachandran score for a single residue, including its
    name, ID, observed phi, psi, omega angles, Ramachandran score percentage and
    classification.
    """

    res_name: str
    res_id: int
    chain_id: str
    resname_tag: str
    angles: dict[str, float]
    pct: float
    classification: ConformerClass

    @staticmethod
    def from_phi_psi_omega(
        phi: float,
        psi: float,
        omega: float,
        res_id: int,
        res_name: str,
        chain_id: str,
        resname_tag: str,
    ) -> "ResidueRamaScore":
        result = _check_rama(phi, psi, resname_tag)
        return ResidueRamaScore(
            res_id=res_id,
            res_name=res_name,
            resname_tag=resname_tag,
            chain_id=chain_id,
            **asdict(result),
            angles={"phi": phi, "psi": psi, "omega": omega},
        )


def _classify_phi_psi(
    atom_array: struc.AtomArray,
) -> list["ResidueRamaScore"]:
    """
    Create RamaScore from a protein structure.

    Parameters
    ----------
    atom_array : struc.AtomArray
        The structure.

    Returns
    -------
    list[RamaScore]
        An instance of RamaScore containing the Ramachandran scores for the structure.
    """
    if all(atom_array.hetero):
        warnings.warn(
            "RotamerScore: Cannot be computed; atom_array must contain some protein residues"
        )
        return []
    atom_array = atom_array[
        struc.filter_canonical_amino_acids(atom_array) & ~atom_array.hetero
    ]
    if atom_array.array_length == 0:
        warnings.warn(
            "RamaScore: Cannot be computed; atom_array must contain at least one residue"
        )
        return []

    rama_scores = []
    for chain_id in struc.get_chains(atom_array):
        chain_arr = atom_array[atom_array.chain_id == chain_id]
        phis, psis, omegas, rama_types = _assign_rama_types(chain_arr)
        res_ids, res_names = struc.get_residues(chain_arr)
        # Identify breakpoints in backbone chains; this is to be skipped as
        # their scores are not reliable
        gaps = _identify_gaps_in_atom_array(chain_arr)

        for phi, psi, omega, res_id, res_name, rama_type in zip(
            phis,
            psis,
            omegas,
            res_ids[1:-1],
            res_names[1:-1],
            rama_types,
        ):
            if res_id in gaps:
                # Skip residues with gaps in the backbone chain
                continue
            if res_name not in RotamerGridResidueMap.__members__:
                # Skip residues that are not in the grid map
                continue

            rama_scores.append(
                ResidueRamaScore.from_phi_psi_omega(
                    phi=phi,
                    psi=psi,
                    omega=omega,
                    res_id=res_id,
                    res_name=res_name,
                    chain_id=atom_array.chain_id[atom_array.res_id == res_id][0],
                    resname_tag=rama_type,
                )
            )

    return rama_scores


@dataclass
class RamaScore:
    """
    Ramachandran score for a given protein structure.
    This class represents calculates the likelihood of each residue's phi/psi angles to occur in crystallized structures
    based on the Top8000 Ramachandran contour grids. It includes their names, IDs, observed phi, psi, omega angles,
    Ramachandran score percentages, and classifications. The thresholsds for classification are defined as follows:
    - FAVORED: pct >= `_RAMA_FAVORED_THRESHOLD`
    - ALLOWED: pct >= `_RAMA_ALLOWED_THRESHOLD` (or `_RAMA_GENERAL_ALLOWED_THRESHOLD` for "general" residues,
    or `_RAMA_CISPRO_ALLOWED_THRESHOLD` for "cispro" residues)
    - OUTLIER: pct < `_RAMA_ALLOWED_THRESHOLD` (or `_RAMA_GENERAL_ALLOWED_THRESHOLD` for "general" residues,
    or `_RAMA_CISPRO_ALLOWED_THRESHOLD` for "cispro" residues)

    It is used to assess the quality of the protein structure based on the Ramachandran angles
    and to identify potential outliers in the phi/psi angles..

    Attributes
    ----------
    rama_scores : list[ResidueRamaScore]
        A list of ResidueRamaScore objects, each representing the Ramachandran score for a residue in the structure.

    References
    ----------
    .. [1] https://pmc.ncbi.nlm.nih.gov/articles/PMC4983197/
    """

    rama_scores: list[ResidueRamaScore] = field(
        default_factory=list,
        metadata={
            "description": "List of Ramachandran scores for each residue in the structure."
        },
    )

    @staticmethod
    def from_atoms(atom_array: struc.AtomArray) -> "RamaScore":
        """
        Create RamaScore from a protein structure.

        Parameters
        ----------
        atom_array : struc.AtomArray
            The structure.

        Returns
        -------
        RamaScore
            An instance of RamaScore containing the Ramachandran scores for the
            structure.
        """
        return RamaScore(rama_scores=_classify_phi_psi(atom_array))


def get_fraction_of_rotamer_outliers(atom_array: struc.AtomArray) -> float:
    """
    Compute the fraction of rotamer outliers for given structure.

    Parameters
    ----------
    atom_array : struc.AtomArray
        The structure.

    Returns
    -------
    float
        The fraction of rotamer outliers in the structure, calculated as the number of outlier
        rotamers divided by the total number of rotamers.
    """
    outlier_rotamers = 0
    total_rotamers = 0
    result = RotamerScore.from_atoms(atom_array)
    for rotamer_score in result.rotamer_scores:
        if rotamer_score.classification == ConformerClass.OUTLIER:
            outlier_rotamers += 1
        total_rotamers += 1
    return outlier_rotamers / total_rotamers if total_rotamers > 0 else 0.0


def get_fraction_of_rama_outliers(
    atom_array: struc.AtomArray,
) -> float:
    """
    Compute the fraction of Ramachandran outliers for given structure.

    Parameters
    ----------
    atom_array : struc.AtomArray
        The AtomArray or AtomArrayStack containing the structure(s).

    Returns
    -------
    float
        The fraction of Ramachandran outliers in the structure, calculated as the number of outlier
        Ramachandran angles divided by the total number of Ramachandran angles.
    """
    outlier_rama = 0
    total_rama = 0
    result = RamaScore.from_atoms(atom_array)

    for rama_score in result.rama_scores:
        if rama_score.classification == ConformerClass.OUTLIER:
            outlier_rama += 1
        total_rama += 1
    return outlier_rama / total_rama if total_rama > 0 else 0.0



__all__ = ["is_small_molecule", "standardize"]


import biotite.structure as struc

DONOR_PATTERN = (
    "["
    "$([Nv3!H0,Nv4!H0+1,nH1]),"
    # Guanidine can be tautomeric - e.g. Arginine
    "$([NX3,NX2]([!O,!S])!@C(!@[NX3,NX2]([!O,!S]))!@[NX3,NX2]([!O,!S])),"
    "$([O,S;!H0])"
    "]"
)
ACCEPTOR_PATTERN = (
    "["
    # Oxygens and Sulfurs
    # singly protonotated can be acceptors
    "$([O,S;v2H1]),"
    # O,S that is unprotonotated, neutral or negative (but not part of nitro-like group!)
    "$([O,S;v2H0;!$([O,S]=N-*)]),"
    "$([O,S;-;!$(*-N=[O,S])]),"
    # also include neutral aromatic oxygen and sulfur
    "$([s,o;+0]),"
    # Nitrogens
    # aromatic unprotonated nitrogens (not trivalent connectivity?)
    "$([nH0+0;!X3]),"
    # nitrile
    "$([ND1H0;$(N#[Cv4])]),"
    # unprotonated nitrogen next to aromatic ring
    "$([Nv3H0;$(N-c)]),"
    # Fluorine on aromatic ring, only
    "$([F;$(F-[#6]);!$(FC[F,Cl,Br,I])])"
    "]"
)
HALOGEN_PATTERN = "[F,Cl,Br,I;+0]"
HBOND_DISTANCE_SCALING = (0.8, 1.15)
HALOGEN_DISTANCE_SCALING = (0.8, 1.05)


def is_small_molecule(chain: struc.AtomArray) -> bool:
    """
    Check whether the given chain is a small molecule.

    Parameters
    ----------
    chain : struc.AtomArray, shape=(n,)
        The chain to check.

    Returns
    -------
    bool
        Whether the chain is a small molecule.
    """
    return chain.hetero[0].item()


def standardize(
    system: struc.AtomArray | struc.AtomArrayStack,
    remove_monoatomic_ions: bool = True,
) -> struc.AtomArray | struc.AtomArrayStack:
    """
    Standardize the given system.

    This function

    - removes hydrogen atoms
    - removes solvent atoms
    - removes monoatomic ions, if specified
    - checks if an associated :class:`biotite.structure.BondList` is present

    Parameters
    ----------
    system : struc.AtomArray or struc.AtomArrayStack
        The system to standardize.
    remove_monoatomic_ions : bool, optional
        If set to ``True``, monoatomic ions will be removed from system.

    Returns
    -------
    struc.AtomArray or struc.AtomArrayStack
        The standardized system.
    """
    if system.bonds is None:
        raise ValueError("The system must have an associated BondList")
    mask = (system.element != "H") & ~struc.filter_solvent(system)
    if remove_monoatomic_ions:
        mask &= ~struc.filter_monoatomic_ions(system)
    return system[..., mask]



__all__ = [
    "GraphMatchWarning",
    "UnmappableEntityError",
    "StructureMismatchError",
    "filter_matched",
    "find_optimal_match",
    "find_all_matches",
    "find_matching_centroids",
]

import itertools
import warnings
from collections import Counter
from collections.abc import Callable
from typing import Any, Iterator
import biotite.interface.rdkit as rdkit_interface
import biotite.sequence as seq
import biotite.sequence.align as align
import biotite.structure as struc
import numpy as np
from numpy.typing import NDArray
from rdkit.Chem import (
    AssignStereochemistryFrom3D,
    BondType,
    Mol,
    SanitizeFlags,
    SanitizeMol,
)
from peppr.common import is_small_molecule

# To match atoms between pose and reference chain,
# the residue and atom name are sufficient to unambiguously identify an atom
_ANNOTATIONS_FOR_ATOM_MATCHING = ["res_name", "atom_name"]
_IDENTITY_MATRIX = align.SubstitutionMatrix(
    seq.ProteinSequence.alphabet,
    seq.ProteinSequence.alphabet,
    np.eye(len(seq.ProteinSequence.alphabet), dtype=np.int32),
)
_PADDING = -1
_UNMAPPABLE_ENTITY_ID = -1


class GraphMatchWarning(UserWarning):
    """
    This warning is raised if the RDKit based molecule matching fails.
    In this case small molecule reordering is skipped.
    """

    pass


class UnmappableEntityError(Exception):
    """
    This exception is raised if the reference and pose structure contain
    entities that cannot be mapped to each other.
    """

    pass


class StructureMismatchError(Exception):
    """
    This exception is raised if the reference and pose structure filtered to
    ``matched`` atoms do not actually match.
    This indicates some issue in the matching process.
    """

    pass


def filter_matched(
    reference: struc.AtomArray | struc.AtomArrayStack,
    pose: struc.AtomArray | struc.AtomArrayStack,
    prefilter: Callable[[struc.AtomArray], NDArray[np.bool_]] | None = None,
) -> tuple[
    struc.AtomArray | struc.AtomArrayStack, struc.AtomArray | struc.AtomArrayStack
]:
    """
    Filter the matched atoms from the reference and pose, i.e.
    where their ``matched`` annotation is ``True``.

    Parameters
    ----------
    reference, pose : AtomArray
        The structures to filter.
        If they have a ``matched`` annotation, the atoms where ``matched`` is ``True``
        are kept.
        Otherwise, all atoms are kept.
    prefilter : Callable[AtomArray -> ndarray, dtype=bool], optional
        For convenience, an additional filter function can be applied to both, the
        reference and pose, before additionally filtering the matched atoms.

    Returns
    -------
    matched_reference, matched_pose : AtomArray or AtomArrayStack
        The filtered reference and pose.

    Raises
    ------
    StructureMismatchError
        If the matched structures do not have corresponding atoms, indicating an issue
        in the prior matching process.
    """
    if "matched" in reference.get_annotation_categories():
        ref_mask = reference.matched
    else:
        ref_mask = np.full(reference.array_length(), True)
    if "matched" in pose.get_annotation_categories():
        pose_mask = pose.matched
    else:
        pose_mask = np.full(pose.array_length(), True)
    if prefilter is not None:
        ref_mask = ref_mask & prefilter(reference)
        pose_mask = pose_mask & prefilter(pose)
    reference = reference[..., ref_mask]
    pose = pose[..., pose_mask]
    if reference.array_length() != pose.array_length():
        raise StructureMismatchError(
            f"Filtered reference has {reference.array_length()} atoms, "
            f"but filtered pose has {pose.array_length()} atoms"
        )
    if np.any(np.char.upper(reference.element) != np.char.upper(pose.element)):
        raise StructureMismatchError(
            "Filtered reference and pose have different chemical elements "
            "mapped to each other"
        )
    return reference, pose


def find_optimal_match(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    min_sequence_identity: float = 0.95,
    use_heuristic: bool = True,
    max_matches: int | None = None,
    allow_unmatched_entities: bool = False,
    use_entity_annotation: bool = False,
    use_structure_match: bool = False,
) -> tuple[struc.AtomArray, struc.AtomArray]:
    """
    Match the atoms from the given reference and pose structure so that the RMSD between
    them is minimized.

    'Matching' has two effects here:

    - Chains and atoms within each residue *that have a counterpart* in the respective
      other structure, are reordered if necessary so that they are in the same order.
    - A ``matched`` annotation is added, which is ``False`` for all atoms, that
      *do not have a counterpart*.

    Parameters
    ----------
    reference : AtomArray, shape=(p,)
        The reference structure.
    pose : AtomArray, shape=(q,)
        The pose structure.
    min_sequence_identity : float, optional
        The minimum sequence identity between two chains to be considered the same
        entity.
    use_heuristic : bool or int, optional
        Whether to employ a fast heuristic [1]_ to find the optimal chain permutation.
        This heuristic represents each chain by its centroid, i.e. instead of
        exhaustively superimposing all atoms for each permutation, only the centroids
        are superimposed and the closest match between the reference and pose is
        selected.
    max_matches : int, optional
        The maximum number of atom mappings to try, if the `use_heuristic` is set to
        ``False``.
    allow_unmatched_entities : bool, optional
        If set to ``True``, allow entire entities to be unmatched.
        This is useful if a pose is compared to a reference which may contain different
        molecules.
    use_entity_annotation : bool, optional
        If set to ``True``, use the ``entity_id`` annotation to determine which chains
        are the same entity and therefore are mappable to each other.
        By default, the entity is determined from sequence identity for polymers and
        residue name for small molecules.
    use_structure_match : bool, optional
        If set to ``True``, use structure matching, i.e. isomorphism on the bond graph,
        to determine which small molecules are the same entity.
        Otherwise, match small molecules by residue name.
        Note that the structure match requires more computation time.

    Returns
    -------
    matched_reference, matched_pose : AtomArray, shape=(p,) or (q,)
        The input atoms, where the chains and atoms within each residue are brought into
        the corresponding order.
        Atoms that are matched between the reference and the pose are annotated with
        ``matched=True``.
        All other atoms are annotated with ``matched=False``.
        This means indexing both structures with ``matched`` as boolean mask will return
        structures with the same number of atoms.

    Notes
    -----
    Atoms that are not matched (``matched=False``), are positioned in the reordered
    return value as follows:

    - Unmatched chains are appended to the end.
    - Unmatched residues within a matched chain are kept at their original sequence
      position.
    - Unmatched atoms within a matched residue are kept at their original position.

    Note that the heuristic used by default is much faster compared to the
    exhaustive approach:
    Especially for larger complexes with many homomers or small molecule copies,
    the number of possible mappings combinatorially explodes.
    However, the heuristic might not find the optimal permutation for all cases,
    especially in poses that only remotely resemble the reference.

    References
    ----------
    .. [1] *Protein complex prediction with AlphaFold-Multimer*, Section 7.3, https://doi.org/10.1101/2021.10.04.463034
    """
    reference_chains = list(struc.chain_iter(reference))
    pose_chains = list(struc.chain_iter(pose))
    if len(reference_chains) == 0:
        raise UnmappableEntityError("Reference is empty")
    if len(pose_chains) == 0:
        raise UnmappableEntityError("Pose is empty")

    if use_heuristic:
        return _find_optimal_match_fast(
            reference_chains,
            pose_chains,
            min_sequence_identity,
            allow_unmatched_entities,
            use_entity_annotation,
            use_structure_match,
        )
    else:
        return _find_optimal_match_precise(
            reference_chains,
            pose_chains,
            min_sequence_identity,
            max_matches,
            allow_unmatched_entities,
            use_entity_annotation,
            use_structure_match,
        )


def find_all_matches(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    min_sequence_identity: float = 0.95,
    allow_unmatched_entities: bool = False,
    use_entity_annotation: bool = False,
    use_structure_match: bool = False,
) -> Iterator[tuple[struc.AtomArray, struc.AtomArray]]:
    """
    Find all possible atom mappings between the reference and the pose.

    Each mappings gives corresponding atoms between the reference and the pose.

    Parameters
    ----------
    reference : AtomArray, shape=(p,)
        The reference structure.
    pose : AtomArray, shape=(q,)
        The pose structure.
    min_sequence_identity : float, optional
        The minimum sequence identity between two chains to be considered the same
        entity.
    allow_unmatched_entities : bool, optional
        If set to ``True``, allow entire entities to be unmatched.
        This is useful if a pose is compared to a reference which may contain different
        molecules.
    use_entity_annotation : bool, optional
        If set to ``True``, use the ``entity_id`` annotation to determine which chains
        are the same entity and therefore are mappable to each other.
        By default, the entity is determined from sequence identity for polymers and
        residue name for small molecules.
    use_structure_match : bool, optional
        If set to ``True``, use structure matching, i.e. isomorphism on the bond graph,
        to determine which small molecules are the same entity.
        Otherwise, match small molecules by residue name.
        Note that the structure match requires more computation time.

    Yields
    ------
    list of tuple (AtomArray, shape=(p,), AtomArray, shape=(q,))
        The input atoms, where the chains and atoms within each chain are brought into
        the corresponding order.
        Atoms that are matched between the reference and the pose are annotated with
        ``matched=True``.
        All other atoms are annotated with ``matched=False``.
        This means indexing both structures with ``matched`` as boolean mask will return
        structures with the same number of atoms.

    See Also
    --------
    find_optimal_match : More detailed information.

    Notes
    -----
    This functions tries all chain mappings of chain that are the same entity and
    within each small molecules tries all proper molecule permutations.
    For larger homomers this can quickly lead to a combinatorial explosion.
    """
    reference_chains = list(struc.chain_iter(reference))
    pose_chains = list(struc.chain_iter(pose))
    for m in _all_global_mappings(
        reference_chains,
        pose_chains,
        min_sequence_identity,
        allow_unmatched_entities,
        use_entity_annotation,
        use_structure_match,
    ):
        yield m


def find_matching_centroids(
    reference_centroids: NDArray[np.floating],
    pose_centroids: NDArray[np.floating],
    reference_entity_ids: NDArray[np.int_] | None = None,
    pose_entity_ids: NDArray[np.int_] | None = None,
) -> tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Greedily find pairs of chains (each represented by its centroid) between the
    reference and the pose that are closest to each other.

    This functions iteratively chooses pairs with the smallest centroid distance, i.e.
    first the pair with the smallest centroid distance is chosen, then the pair with the
    second smallest centroid distance and so on.

    Parameters
    ----------
    reference_centroids, pose_centroids : np.ndarray, shape=(n,3)
        The centroids of the reference and pose chains.
    reference_entity_ids, pose_entity_ids : np.ndarray, shape=(n,), dtype=int, optional
        The entity IDs of the chains.
        Only centroids of chains with the same entity ID can be matched.
        By default, all can be matched to each other.

    Returns
    -------
    reference_chain_indices, pose_chain_indices : np.ndarray, shape=(n,)
        Indices to corresponding chains in the reference and pose that gives the pairs
        with the smallest distance to each other.
    """
    if reference_entity_ids is None or pose_entity_ids is None:
        # Assign the same entity id to all chains
        reference_entity_ids = np.zeros(len(reference_centroids), dtype=int)
        pose_entity_ids = np.zeros(len(pose_centroids), dtype=int)

    distances = struc.distance(reference_centroids[:, None], pose_centroids[None, :])
    # Different entities must not be matched
    distances[reference_entity_ids[:, None] != pose_entity_ids[None, :]] = np.inf
    # Unmappable entities must not be matched
    distances[reference_entity_ids == _UNMAPPABLE_ENTITY_ID, :] = np.inf
    distances[:, pose_entity_ids == _UNMAPPABLE_ENTITY_ID] = np.inf

    reference_chain_indices = []
    pose_chain_indices = []
    while True:
        min_distance = np.min(distances)
        if np.isinf(min_distance):
            # No chains can be matched to each other anymore
            break
        min_reference_i, min_pose_i = np.argwhere(distances == min_distance)[0]
        reference_chain_indices.append(min_reference_i)
        pose_chain_indices.append(min_pose_i)
        distances[min_reference_i, :] = np.inf
        distances[:, min_pose_i] = np.inf
    reference_chain_indices = np.array(reference_chain_indices)
    pose_chain_indices = np.array(pose_chain_indices)
    # Try to keep the reference chains in the original order if possible
    order = np.argsort(reference_chain_indices)
    reference_chain_indices = reference_chain_indices[order]
    pose_chain_indices = pose_chain_indices[order]
    return reference_chain_indices, pose_chain_indices


def _find_optimal_match_fast(
    reference_chains: list[struc.AtomArray],
    pose_chains: list[struc.AtomArray],
    min_sequence_identity: float,
    allow_unmatched_entities: bool,
    use_entity_annotation: bool,
    use_structure_match: bool,
) -> tuple[struc.AtomArray, struc.AtomArray]:
    """
    Find matching atoms that minimize the centroid RMSD between the pose and the
    reference.

    Parameters
    ----------
    reference_chains, pose_chains : list of AtomArray
        The reference and pose structure, separated into chains.
    min_sequence_identity : float, optional
        The minimum sequence identity between two chains to be considered the same
        entity.
    allow_unmatched_entities : bool, optional
        If set to ``True``, allow entire entities to be unmatched.
        This is useful if a pose is compared to a reference which may contain different
        molecules.
    use_entity_annotation : bool, optional
        If set to ``True``, use the ``entity_id`` annotation to determine which chains
        are the same entity and therefore are mappable to each other.
    use_structure_match : bool, optional
        If set to ``True``, use structure matching, i.e. isomorphism on the bond graph,
        to determine which small molecules are the same entity.
        Otherwise, match small molecules by residue name.
        Note that the structure match requires more computation time.

    Returns
    -------
    matched_reference, matched_pose : AtomArray
        The input atoms, where the chains and atoms within each chain are brought into
        the corresponding order.
    """
    reference_entity_ids, pose_entity_ids = _assign_entity_ids(
        reference_chains,
        pose_chains,
        min_sequence_identity,
        allow_unmatched_entities,
        use_entity_annotation,
        use_structure_match,
    )

    # Find corresponding chains by identifying the chain permutation that minimizes
    # the centroid RMSD
    reference_centroids = np.array([struc.centroid(c) for c in reference_chains])
    pose_centroids = np.array([struc.centroid(c) for c in pose_chains])
    best_transform = None
    best_rmsd = np.inf
    best_reference_indices = None
    best_pose_indices = None
    # Test all possible chains that represent the same entity against the anchor chain
    for ref_i, pose_i in _all_anchor_combinations(
        reference_chains, reference_entity_ids, pose_chains, pose_entity_ids
    ):
        # Superimpose the entire system
        # based on the anchor and chosen reference chain
        transform = _get_superimposition_transform(
            reference_chains[ref_i], pose_chains[pose_i]
        )
        superimposed_pose_centroids = transform.apply(pose_centroids)
        reference_indices, pose_indices = find_matching_centroids(
            reference_centroids,
            superimposed_pose_centroids,
            reference_entity_ids,
            pose_entity_ids,
        )
        rmsd = struc.rmsd(
            reference_centroids[reference_indices],
            superimposed_pose_centroids[pose_indices],
        )
        if rmsd < best_rmsd:
            best_rmsd = rmsd
            best_transform = transform
            best_reference_indices = reference_indices
            best_pose_indices = pose_indices

    pose_chains = [best_transform.apply(chain) for chain in pose_chains]  # type: ignore[union-attr]
    return _match_using_chain_order(
        reference_chains,
        pose_chains,
        best_reference_indices,
        best_pose_indices,
        # Superimposition is already defined by at least two centroids,
        # as they leave no degrees of freedom
        superimpose=False if len(reference_centroids) > 1 else True,
    )


def _find_optimal_match_precise(
    reference_chains: list[struc.AtomArray],
    pose_chains: list[struc.AtomArray],
    min_sequence_identity: float,
    max_matches: int | None,
    allow_unmatched_entities: bool,
    use_entity_annotation: bool,
    use_structure_match: bool,
) -> tuple[struc.AtomArray, struc.AtomArray]:
    """
    Find matching atoms that minimize that minimize the all-atom RMSD between the pose
    and the reference.

    Parameters
    ----------
    reference_chains, pose_chains : list of AtomArray
        The reference and pose structure, separated into chains.
    min_sequence_identity : float
        The minimum sequence identity between two chains to be considered the same
        entity.
    max_matches : int, optional
        The maximum number of mappings to try.
    allow_unmatched_entities : bool, optional
        If set to ``True``, allow entire entities to be unmatched.
        This is useful if a pose is compared to a reference which may contain different
        molecules.
    use_entity_annotation : bool, optional
        If set to ``True``, use the ``entity_id`` annotation to determine which chains
        are the same entity and therefore are mappable to each other.
    use_structure_match : bool, optional
        If set to ``True``, use structure matching, i.e. isomorphism on the bond graph,
        to determine which small molecules are the same entity.
        Otherwise, match small molecules by residue name.
        Note that the structure match requires more computation time.

    Returns
    -------
    matched_reference, matched_pose : AtomArray
        The input atoms, where the chains and atoms within each chain are brought into
        the corresponding order.
    """
    if max_matches is not None and max_matches < 1:
        raise ValueError("Maximum number of mappings must be at least 1")

    best_rmsd = np.inf
    best_mapping: tuple[struc.AtomArray, struc.AtomArray] | None = None
    for it, (mapped_reference, mapped_pose) in enumerate(
        _all_global_mappings(
            reference_chains,
            pose_chains,
            min_sequence_identity,
            allow_unmatched_entities,
            use_entity_annotation,
            use_structure_match,
        )
    ):
        if max_matches is not None and it >= max_matches:
            break
        matched_ref_coord = mapped_reference.coord[mapped_reference.matched]
        matched_pose_coord = mapped_pose.coord[mapped_pose.matched]
        matched_pose_coord, _ = struc.superimpose(matched_ref_coord, matched_pose_coord)
        rmsd = struc.rmsd(matched_ref_coord, matched_pose_coord)
        if rmsd < best_rmsd:
            best_rmsd = rmsd
            best_mapping = (mapped_reference, mapped_pose)

    if best_mapping is None:
        raise UnmappableEntityError(
            "No chain in the pose can be mapped to a reference chain"
        )
    return best_mapping


def _all_global_mappings(
    reference_chains: list[struc.AtomArray],
    pose_chains: list[struc.AtomArray],
    min_sequence_identity: float = 0.95,
    allow_unmatched_entities: bool = False,
    use_entity_annotation: bool = False,
    use_structure_match: bool = False,
) -> Iterator[tuple[struc.AtomArray, struc.AtomArray]]:
    """
    Find all possible atom mappings between the reference and the pose.

    Each mappings gives corresponding atoms between the reference and the pose.

    Parameters
    ----------
    reference_chains, pose_chains : list of AtomArray
        The reference and pose chains, respectively.
    min_sequence_identity : float
        The minimum sequence identity between two chains to be considered the same
        entity.
    allow_unmatched_entities : bool, optional
        If set to ``True``, allow entire entities to be unmatched.
        This is useful if a pose is compared to a reference which may contain different
        molecules.
    use_entity_annotation : bool, optional
        If set to ``True``, use the ``entity_id`` annotation to determine which chains
        are the same entity and therefore are mappable to each other.
    use_structure_match : bool, optional
        If set to ``True``, use structure matching, i.e. isomorphism on the bond graph,
        to determine which small molecules are the same entity.
        Otherwise, match small molecules by residue name.
        Note that the structure match requires more computation time.

    Yields
    ------
    matched_reference, matched_pose : list of AtomArray
        The input chains, where the chains and atoms within each chain are brought into
        the corresponding order.

    Notes
    -----
    This functions tries all chain mappings of chain that are the same entity and
    within each small molecules tries all proper molecule permutations.
    """
    # Add the `matched` annotation in-place without copying
    # This is OK, as the chain `AtomArray`s were created in private
    # functions, so no user-provided data is modified
    for chain in itertools.chain(reference_chains, pose_chains):
        # Every atom is initialized as unmatched
        chain.set_annotation("matched", np.full(chain.array_length(), False))

    reference_entity_ids, pose_entity_ids = _assign_entity_ids(
        reference_chains,
        pose_chains,
        min_sequence_identity,
        allow_unmatched_entities,
        use_entity_annotation,
        use_structure_match,
    )
    for ref_chain_indices, pose_chain_indices in _all_chain_mappings(
        reference_entity_ids, pose_entity_ids
    ):
        # All possible intra-chain atom mappings for the current chain mapping
        # In case of symmetric small molecules, multiple mappings are possible
        # The outer list iterates over the matched chains
        # The inner list iterates over possible atom mappings within each chain
        all_intra_chain_mappings: list[
            list[tuple[struc.AtomArray, struc.AtomArray]]
        ] = []
        for ref_chain_i, pose_chain_i in zip(ref_chain_indices, pose_chain_indices):
            if is_small_molecule(reference_chains[ref_chain_i]):
                pose_mappings = _molecule_mappings(
                    reference_chains[ref_chain_i], pose_chains[pose_chain_i]
                )
                all_mapping_possibilities = []
                for pose_mapping in pose_mappings:
                    mapped_ref_chain = reference_chains[ref_chain_i].copy()
                    mapped_pose_chain = pose_chains[pose_chain_i][pose_mapping]
                    mapped_ref_chain.matched[:] = True
                    mapped_pose_chain.matched[:] = True
                    all_mapping_possibilities.append(
                        (mapped_ref_chain, mapped_pose_chain)
                    )
                all_intra_chain_mappings.append(all_mapping_possibilities)
            else:
                # For polymers there is only one mapping
                all_intra_chain_mappings.append(
                    [
                        _match_common_residues(
                            # Copy, as the `_match_common_residues()` modifies the
                            # `matched` annotation, so subsequent iteration would
                            # use a modified `matched` annotation
                            reference_chains[ref_chain_i].copy(),
                            pose_chains[pose_chain_i].copy(),
                        )
                    ]
                )

        unmapped_ref_chains = [
            reference_chains[i]
            for i in range(len(reference_chains))
            if i not in ref_chain_indices
        ]
        unmapped_pose_chains = [
            pose_chains[i]
            for i in range(len(pose_chains))
            if i not in pose_chain_indices
        ]
        # Create Cartesian product of intra-chain mappings over all chains
        for mapped_chains in itertools.product(*all_intra_chain_mappings):
            mapped_ref_chains = [chain for chain, _ in mapped_chains]
            mapped_pose_chains = [chain for _, chain in mapped_chains]
            yield (
                struc.concatenate(mapped_ref_chains + unmapped_ref_chains),
                struc.concatenate(mapped_pose_chains + unmapped_pose_chains),
            )


def _all_chain_mappings(
    reference_entity_ids: NDArray[np.int_],
    pose_entity_ids: NDArray[np.int_],
) -> Iterator[tuple[NDArray[np.int_], NDArray[np.int_]]]:
    """
    Iterate over all possible mappings of chains between the reference and the pose
    that are the same entity.

    This function tries to map as many chains as possible, but if either structure
    contains more chains than the other for the same entity, chains may be unmapped
    (i.e. missing) in the returned mappings.

    Parameters
    ----------
    reference_entity_ids, pose_entity_ids : np.ndarray, shape=(n,), dtype=int
        The entity IDs of the chains.

    Yields
    ------
    reference_chain_indices, pose_chain_indices : np.ndarray, shape=(n,), dtype=int
        The indices of the chains in the reference and pose that are mapped to each
        other.
    """
    unique_entity_ids = np.unique(
        np.concatenate([reference_entity_ids, pose_entity_ids])
    )
    unique_entity_ids = unique_entity_ids[unique_entity_ids != _UNMAPPABLE_ENTITY_ID]
    if len(unique_entity_ids) == 0:
        raise UnmappableEntityError(
            "No chain in the pose can be mapped to a reference chain"
        )
    mappings_for_all_entities = []
    for entity_id in unique_entity_ids:
        # Find all chains that belong to the same entity
        ref_chain_indices = np.where(reference_entity_ids == entity_id)[0]
        pose_chain_indices = np.where(pose_entity_ids == entity_id)[0]
        # Pad the shorter index array
        max_length = max(len(ref_chain_indices), len(pose_chain_indices))
        padded_ref_chain_indices = np.full(max_length, _PADDING, dtype=int)
        padded_pose_chain_indices = np.full(max_length, _PADDING, dtype=int)
        padded_ref_chain_indices[: len(ref_chain_indices)] = ref_chain_indices
        padded_pose_chain_indices[: len(pose_chain_indices)] = pose_chain_indices
        ref_chain_indices = padded_ref_chain_indices
        pose_chain_indices = padded_pose_chain_indices
        # Create all possible mappings for chains of this entity
        # Mapping to the padding value means that the corresponding chain is not mapped
        possible_mappings_within_entity = []
        for permutated_pose_chain_indices in itertools.permutations(pose_chain_indices):
            possible_mappings_within_entity.append(
                np.stack([ref_chain_indices, permutated_pose_chain_indices], axis=1)
            )
        mappings_for_all_entities.append(possible_mappings_within_entity)

    # Create Cartesian product of all mappings
    for mapping_for_all_entities in itertools.product(*mappings_for_all_entities):
        mapping = np.concatenate(mapping_for_all_entities, axis=0)
        mapping = mapping[np.all(mapping != _PADDING, axis=1)]
        yield tuple(mapping.T)


def _match_using_chain_order(
    reference_chains: list[struc.AtomArray],
    pose_chains: list[struc.AtomArray],
    reference_chain_order: NDArray[np.int_],
    pose_chain_order: NDArray[np.int_],
    superimpose: bool = False,
) -> tuple[struc.AtomArray, struc.AtomArray]:
    """
    Given the order of corresponding chains between the reference and the pose,
    find the corresponding atom order within each chain and create structures with
    corresponding chain order and the ``matched`` annotation array.

    Parameters
    ----------
    reference_chains, pose_chains : list of AtomArray
        The reference and pose chains, respectively.
    reference_chain_order, pose_chain_order : np.ndarray, shape=(n,), dtype=int
        The indices when applied to `reference_chains` and `pose_chains`, respectively,
        yield the corresponding chains.
    superimpose : bool, optional
        Whether to superimpose the corresponding small molecules onto each other
        before finding the optimal permutation.

    Returns
    -------
    matched_reference, matched_pose: AtomArray
        The reference and pose, respectively, with the chains and atoms in the
        corresponding order.
        The structures have a ``matched`` annotation array, that is ``True`` for
        each atom that has a counterpart in the other structure.
    """
    # Add the `matched` annotation in-place without copying
    # This is OK, as the chain `AtomArray`s were created in private
    # functions, so no user-provided data is modified
    for chain in itertools.chain(reference_chains, pose_chains):
        # Every atom is initialized as unmatched
        chain.set_annotation("matched", np.full(chain.array_length(), False))

    handled_reference_chain_mask = np.full(len(reference_chains), False, dtype=bool)
    handled_pose_chain_mask = np.full(len(pose_chains), False, dtype=bool)
    matched_reference_chains = []
    matched_pose_chains = []
    for ref_i, pose_i in zip(reference_chain_order, pose_chain_order):
        if is_small_molecule(reference_chains[ref_i]):
            ref_atom_order, pose_atom_order = _find_optimal_molecule_permutation(
                reference_chains[ref_i],
                pose_chains[pose_i],
                superimpose=superimpose,
            )
            ref_chain = reference_chains[ref_i][ref_atom_order]
            pose_chain = pose_chains[pose_i][pose_atom_order]
            ref_chain.matched[:] = True
            pose_chain.matched[:] = True
            matched_reference_chains.append(ref_chain)
            matched_pose_chains.append(pose_chain)
        else:
            ref_chain, pose_chain = _match_common_residues(
                reference_chains[ref_i], pose_chains[pose_i]
            )
            matched_reference_chains.append(ref_chain)
            matched_pose_chains.append(pose_chain)
        handled_reference_chain_mask[ref_i] = True
        handled_pose_chain_mask[pose_i] = True

    return (
        struc.concatenate(
            matched_reference_chains
            # The unmatched chains are appended to the end
            + [reference_chains[i] for i in np.where(~handled_reference_chain_mask)[0]]
        ),
        struc.concatenate(
            matched_pose_chains
            # The unmatched chains are appended to the end
            + [pose_chains[i] for i in np.where(~handled_pose_chain_mask)[0]]
        ),
    )


def _match_common_residues(
    reference: struc.AtomArray, pose: struc.AtomArray
) -> tuple[struc.AtomArray, struc.AtomArray]:
    """
    Find common residues (and the common atoms) in two protein chains.

    Parameters
    ----------
    reference, pose : struc.AtomArray, shape=(n,)
        The reference and pose polymer chain whose common atoms are to be found.

    Returns
    -------
    matched_reference, matched_pose : struc.AtomArray, shape=(n,)
        Reordered versions of the input structures, where the common atoms are in the
        corresponding order and unique atoms in each structure get annotated with
        ``matched=False``.
    """
    # Shortcut if the structures already match perfectly atom-wise
    if _is_matched(reference, pose, _ANNOTATIONS_FOR_ATOM_MATCHING):
        reference.matched[:] = True
        pose.matched[:] = True
        return reference, pose

    reference_sequence = struc.to_sequence(reference)[0][0]
    pose_sequence = struc.to_sequence(pose)[0][0]
    alignment = align.align_optimal(
        reference_sequence,
        pose_sequence,
        _IDENTITY_MATRIX,
        # We get mismatches due to cropping, not due to evolution
        # -> linear gap penalty makes most sense
        gap_penalty=-1,
        terminal_penalty=False,
        max_number=1,
    )[0]
    # Remove gaps -> crop structures to common residues
    alignment.trace = alignment.trace[(alignment.trace != -1).all(axis=1)]

    # Atom masks that are True for atoms in residues that are common in both structures
    ref_aligned_mask = _get_mask_from_alignment_trace(reference, alignment.trace[:, 0])
    pose_aligned_mask = _get_mask_from_alignment_trace(pose, alignment.trace[:, 1])

    # Within the atoms of aligned residues, select only common atoms
    ref_order, pose_order, ref_common_mask, pose_common_mask = _find_atom_intersection(
        reference[ref_aligned_mask], pose[pose_aligned_mask]
    )
    reference[ref_aligned_mask] = reference[ref_aligned_mask][ref_order]
    pose[pose_aligned_mask] = pose[pose_aligned_mask][pose_order]
    reference.matched[ref_aligned_mask] = ref_common_mask
    pose.matched[pose_aligned_mask] = pose_common_mask

    return reference, pose


def _get_mask_from_alignment_trace(
    chain: struc.AtomArray, trace_column: NDArray[np.int_]
) -> NDArray[np.bool_]:
    """
    Get a mask that is True for all atoms, whose residue is contained in the given
    alignment trace column.

    Parameters
    ----------
    chain : AtomArray, shape=(n,)
        The chain to get the mask for.
    trace_column : ndarray, shape=(k,), dtype=int
        The column of the alignment trace for that chain.
        Each index in this trace column points to a residue in the chain.

    Returns
    -------
    mask : ndarray, shape=(n,), dtype=bool
        The mask for the given chain.
    """
    return struc.get_residue_masks(chain, struc.get_residue_starts(chain))[
        trace_column
    ].any(axis=0)


def _find_atom_intersection(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
) -> tuple[NDArray[np.int_], NDArray[np.int_], NDArray[np.bool_], NDArray[np.bool_]]:
    """
    Find the intersection of two structures, i.e. the set of equivalent atoms.

    Parameters
    ----------
    reference, pose : AtomArray
        The reference and pose chain, respectively.

    Returns
    -------
    ref_order, pose_order : ndarray, shape=(n,), dtype=int
        The reference and pose indices that bring the input structures into a common
        atom order.
    intersection_ref_mask, intersection_pose_mask : ndarray, shape=(n,), dtype=bool
        The mask that is `True` for each atom in the sorted atom order, that appears
        in both structures.

    Notes
    -----
    The order is not necessarily canonical, as the atoms are simply sorted based on the
    given annotations.
    The important requirement is that the order is the same for both structures.
    """
    # Shortcut if the structures already match perfectly atom-wise
    if _is_matched(reference, pose, _ANNOTATIONS_FOR_ATOM_MATCHING):
        return (
            np.arange(reference.array_length()),
            np.arange(reference.array_length()),
            np.full(reference.array_length(), True),
            np.full(reference.array_length(), True),
        )

    # Use continuous residue IDs to enforce that the later reordering does not mix up
    # atoms from different residues
    reference.res_id = struc.create_continuous_res_ids(reference, False)
    pose.res_id = struc.create_continuous_res_ids(pose, False)
    # Implicitly expect that the annotation array dtypes are the same for both
    structured_dtype = np.dtype(
        [
            (name, pose.get_annotation(name).dtype)
            for name in ["res_id"] + _ANNOTATIONS_FOR_ATOM_MATCHING
        ]
    )
    ref_annotations = _annotations_to_structured(reference, structured_dtype)
    pose_annotations = _annotations_to_structured(pose, structured_dtype)
    # Atom ordering might not be same -> sort
    ref_order = np.argsort(ref_annotations)
    pose_order = np.argsort(pose_annotations)
    # Identify the intersection of the two annotation arrays
    intersection_ref_mask = np.isin(ref_annotations[ref_order], pose_annotations)
    intersection_pose_mask = np.isin(pose_annotations[pose_order], ref_annotations)

    return ref_order, pose_order, intersection_ref_mask, intersection_pose_mask


def _annotations_to_structured(
    atoms: struc.AtomArray, structured_dtype: np.dtype
) -> NDArray[Any]:
    """
    Convert atom annotations into a single structured `ndarray`.

    Parameters
    ----------
    atoms : AtomArray
        The annotation arrays are taken from this structure.
    structured_dtype : dtype
        The dtype of the structured array to be created.
        The fields of the dtype determine which annotations are taken from `atoms`.
    """
    if structured_dtype.fields is None:
        raise TypeError("dtype must be structured")
    structured = np.zeros(atoms.array_length(), dtype=structured_dtype)
    for field in structured_dtype.fields:
        structured[field] = atoms.get_annotation(field)
    return structured


def _find_optimal_molecule_permutation(
    reference: struc.AtomArray, pose: struc.AtomArray, superimpose: bool
) -> tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Find corresponding atoms in small molecules that minimizes the RMSD between the
    pose and the reference.

    Use graph isomorphism on the bond graph to account for symmetries within
    the small molecule.

    Parameters
    ----------
    reference, pose : struc.AtomArray, shape=(n,)
        The reference and pose small molecule, respectively.
    superimpose : bool
        Whether to superimpose the reference and the reordered pose onto each other
        before calculating the RMSD.
        Only necessary if no other means has already determined the superimposition.
        If ``False``, it is expected that they are already superimposed onto each other.

    Returns
    -------
    reference_order, pose_order : np.array, shape=(n,), dtype=int
        The indices that when applied to the `reference` or `pose`, respectively,
        yield the corresponding atoms.
    """
    best_rmsd = np.inf
    best_pose_atom_order = None
    for pose_atom_order in _molecule_mappings(reference, pose):
        if len(pose_atom_order) != reference.array_length():
            raise ValueError("Atom mapping does not cover all atoms")
        if superimpose:
            superimposed, _ = struc.superimpose(reference, pose[pose_atom_order])
        else:
            superimposed = pose[pose_atom_order]
        rmsd = struc.rmsd(reference, superimposed)
        if rmsd < best_rmsd:
            best_rmsd = rmsd
            best_pose_atom_order = pose_atom_order
    return np.arange(reference.array_length()), best_pose_atom_order  # type: ignore[return-value]


def _molecule_mappings(
    reference: struc.AtomArray, pose: struc.AtomArray
) -> list[NDArray[np.int_]]:
    """
    Find corresponding atoms in small molecules.

    Use graph isomorphism on the bond graph to account for symmetries within
    the small molecule.

    Parameters
    ----------
    reference, pose : struc.AtomArray, shape=(n,)
        The reference and pose small molecule, respectively.
        It is expected that they are already superimposed onto each other.

    Returns
    -------
    list of np.ndarray, shape=(n,), dtype=int
        All mappings of `pose` that maps its atoms to the `reference` with
        respect to elements and the bond graph.
    """
    reference_mol = _to_mol(reference)
    pose_mol = _to_mol(pose)
    mappings = pose_mol.GetSubstructMatches(
        reference_mol, useChirality=True, uniquify=False
    )
    if len(mappings) == 0:
        # The coordinates may be off, so try matching without chirality
        mappings = pose_mol.GetSubstructMatches(
            reference_mol, useChirality=False, uniquify=False
        )
    if len(mappings) == 0:
        # If still no match is found, conformation is not the problem
        # -> Check if bond graph is the problem
        if np.array_equal(reference.element, pose.element):
            # If the elements are the same, probably some simple
            # incompatible bonds are the problem
            # -> Assume that the atom order is the same in reference and pose
            mappings = [np.arange(pose.array_length())]
            warnings.warn(
                "Incompatible bond graph between pose and reference small molecule",
                GraphMatchWarning,
            )
        else:
            # They are probably different incompatible molecules
            # This is a user error, as different molecules should have different
            # residue names
            raise UnmappableEntityError(
                "No atom mapping found between pose and reference small molecule "
                f"'{reference.res_name[0]}'"
            )

    # Convert tuples to proper index arrays
    return [np.asarray(mapping) for mapping in mappings]


def _assign_entity_ids(
    reference_chains: list[struc.AtomArray],
    pose_chains: list[struc.AtomArray],
    min_sequence_identity: float,
    allow_unmatched_entities: bool = False,
    use_entity_annotation: bool = False,
    use_structure_match: bool = False,
) -> tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Assign a unique entity ID to each distinct chain.

    This means that two chains with the same entity ID have sufficient sequence
    identity or in case of small molecules have the same ``res_name`` (or the same
    bond graph if `use_structure_match` is set to ``True``).

    Parameters
    ----------
    reference_chains, pose_chains : list of struc.AtomArray, length=p or length=q
        The reference and pose chains, respectively.
    min_sequence_identity : float
        The minimum sequence identity between two chains to be considered the same
        entity.
    allow_unmatched_entities : bool, optional
        If set to ``True``, allow the reference and pose to have different entities.
        Otherwise an :class:`UnmappableEntityError` is raised.
    use_entity_annotation : bool, optional
        If set to ``True``, use the ``entity_id`` annotation to determine which chains
        are the same entity and therefore are mappable to each other.
    use_structure_match : bool, optional
        If set to ``True``, use structure matching, i.e. isomorphism on the bond graph,
        to determine which small molecules are the same entity.
        Otherwise, match small molecules by residue name.
        Note that the structure match requires more computation time.

    Returns
    -------
    reference_entity_ids, pose_entity_ids : np.ndarray, shape=(p,) or shape=(q,), dtype=int
        The entity IDs.
        Chains that are not mappable between reference and pose get the ID ``-1``.
    """
    # Assign reference and pose entity IDs in a single call
    # in order to assign the same ID to corresponding chains between reference and pose
    entity_ids = _assign_entity_ids_to_chains(
        reference_chains + pose_chains,
        min_sequence_identity,
        use_entity_annotation,
        use_structure_match,
    )

    # Split the entity IDs again
    reference_entity_ids = entity_ids[: len(reference_chains)]
    pose_entity_ids = entity_ids[len(reference_chains) :]
    # In the worst case, the number of distinct entity IDs
    # is equal to the number of chains -> use this as 'minlength'
    reference_entity_id_counts = np.bincount(
        reference_entity_ids, minlength=len(entity_ids)
    )
    pose_entity_id_counts = np.bincount(pose_entity_ids, minlength=len(entity_ids))
    if (
        not allow_unmatched_entities
        and (reference_entity_id_counts != pose_entity_id_counts).any()
    ):
        raise UnmappableEntityError(
            "Reference and pose have different entities or a different multiplicity"
        )
    # Assign an unmappable entity ID to IDs that only appear in one structure
    reference_entity_ids[pose_entity_id_counts[reference_entity_ids] == 0] = (
        _UNMAPPABLE_ENTITY_ID
    )
    pose_entity_ids[reference_entity_id_counts[pose_entity_ids] == 0] = (
        _UNMAPPABLE_ENTITY_ID
    )

    return reference_entity_ids, pose_entity_ids


def _assign_entity_ids_to_chains(
    chains: list[struc.AtomArray],
    min_sequence_identity: float,
    use_entity_annotation: bool,
    use_structure_match: bool,
) -> NDArray[np.int_]:
    """
    Assign a unique entity ID to each distinct chain.

    This means that two chains with the same entity ID have sufficient sequence
    identity or in case of small molecules have the same ``res_name`` (or the same
    bond graph if `use_structure_match` is set to ``True``).

    Parameters
    ----------
    chains : list of struc.AtomArray, length=n
        The chains to assign entity IDs to.
    min_sequence_identity : float
        The minimum sequence identity between two chains to be considered the same
        entity.
    use_entity_annotation : bool, optional
        If set to ``True``, use the ``entity_id`` annotation to determine which chains
        are the same entity and therefore are mappable to each other.
    use_structure_match : bool, optional
        If set to ``True``, use structure matching, i.e. isomorphism on the bond graph,
        to determine which small molecules are the same entity.
        Otherwise, match small molecules by residue name.
        Note that the structure match requires more computation time.


    Returns
    -------
    entity_ids : np.ndarray, shape=(n,), dtype=int
        The entity IDs.
    """
    entity_ids: list[int] = []
    if use_entity_annotation:
        for chain in chains:
            if "entity_id" not in chain.get_annotation_categories():
                raise struc.BadStructureError(
                    "'use_entity_annotation=True' requires the 'entity_id' annotation"
                )
            entity_id = chain.entity_id[0].item()
            if not np.all(chain.entity_id == entity_id):
                raise struc.BadStructureError(
                    "'entity_id' annotation must be the same for all atoms in a chain"
                )
            entity_ids.append(chain.entity_id[0].item())
        return np.array(entity_ids, dtype=int)

    sequences = [
        struc.to_sequence(chain)[0][0] if not is_small_molecule(chain) else None
        for chain in chains
    ]
    if use_structure_match:
        molecules = [
            _to_mol(chain) if is_small_molecule(chain) else None for chain in chains
        ]

    current_entity_id = 0
    for i, (chain, sequence) in enumerate(zip(chains, sequences)):
        for j in range(i):
            if (sequence is None and sequences[j] is not None) or (
                sequence is not None and sequences[j] is None
            ):
                # Cannot match small molecules to polymer chains
                continue
            elif sequence is None:
                # Small molecule case
                if use_structure_match:
                    # It is only a complete structure match,
                    # if i is a non-strict subset of j and j is a non-strict subset of i
                    if (
                        molecules[i].HasSubstructMatch(molecules[j]) and  # type: ignore[union-attr]
                        molecules[j].HasSubstructMatch(molecules[i])  # type: ignore[union-attr]
                    ):  # fmt: skip
                        entity_ids.append(entity_ids[j])
                        break
                else:
                    # Match small molecules by residue name
                    if chain.res_name[0] == chains[j].res_name[
                        0
                    ] and _equal_composition([chain, chains[j]]):
                        entity_ids.append(entity_ids[j])
                        break
            else:
                # Match polymer chains by sequence identity
                alignment = align.align_optimal(
                    sequence,
                    sequences[j],
                    _IDENTITY_MATRIX,
                    # We get mismatches due to experimental artifacts, not evolution
                    # -> linear gap penalty makes most sense
                    gap_penalty=-1,
                    terminal_penalty=False,
                    max_number=1,
                )[0]
                if (
                    align.get_sequence_identity(alignment, mode="shortest")
                    >= min_sequence_identity
                ):
                    entity_ids.append(entity_ids[j])
                    break
        else:
            # No match found to a chain that already has an entity ID -> assign new ID
            entity_ids.append(current_entity_id)
            current_entity_id += 1

    return np.array(entity_ids, dtype=int)


def _all_anchor_combinations(
    reference_chains: list[struc.AtomArray],
    reference_entity_ids: NDArray[np.int_],
    pose_chains: list[struc.AtomArray],
    pose_entity_ids: NDArray[np.int_],
) -> Iterator[tuple[int, int]]:
    """
    Choose the reference and pose anchor chain for the heuristic chain matching.

    The most preferable chain is the one with the least multiplicity and the longest
    sequence.

    Parameters
    ----------
    chains : list of struc.AtomArray, length=n
        The chains to choose from.
    entity_ids : ndarray, shape=(n,), dtype=int
        The entity IDs of the chains.
        Used to determine the multiplicity of each chain.

    Yields
    ------
    reference_i, pose_i : int
        The putative anchor indices.
    """
    mappable_mask = pose_entity_ids != _UNMAPPABLE_ENTITY_ID
    if not mappable_mask.any():
        raise UnmappableEntityError(
            "No chain in the pose can be mapped to a reference chain"
        )
    protein_chain_mask = np.array(
        [not is_small_molecule(chain) for chain in pose_chains]
    )
    valid_anchor_mask = protein_chain_mask & mappable_mask
    if not valid_anchor_mask.any():
        # No mappable protein chains
        # -> Simply use the first mappable small molecule as anchor
        anchor_entity_id = pose_entity_ids[np.where(mappable_mask)[0][0]]
    else:
        valid_anchor_indices = np.where(valid_anchor_mask)[0]
        protein_entity_ids = pose_entity_ids[valid_anchor_indices]
        multiplicities_of_entity_ids = np.bincount(protein_entity_ids)
        multiplicities = multiplicities_of_entity_ids[protein_entity_ids]
        least_multiplicity_indices = np.where(multiplicities == np.min(multiplicities))[
            0
        ]
        # Use the sequence length as tiebreaker
        sequence_lengths = np.array([len(pose_chains[i]) for i in valid_anchor_indices])
        # Only consider the lengths of the preselected chains
        largest_length = np.max(sequence_lengths[least_multiplicity_indices])
        largest_length_indices = np.where(sequence_lengths == largest_length)[0]
        best_pose_anchors = np.intersect1d(
            least_multiplicity_indices, largest_length_indices
        )
        anchor_entity_id = pose_entity_ids[valid_anchor_mask][best_pose_anchors[0]]

    # Check if there is a 1-to-1 correspondence of the anchor entity between
    # the reference and pose
    if np.count_nonzero(reference_entity_ids == anchor_entity_id) == np.count_nonzero(
        pose_entity_ids == anchor_entity_id
    ):
        # It is sufficient to keep the pose anchor fixed and iterate over the reference
        pose_i = np.where(pose_entity_ids == anchor_entity_id)[0][0]
        for ref_i in range(len(reference_chains)):
            if reference_entity_ids[ref_i] == anchor_entity_id:
                yield ref_i, pose_i
    else:
        # Otherwise an anchor chain in the pose might be missing in the reference
        # -> We need to iterate over all pose anchors as well
        for pose_i in np.where(pose_entity_ids == anchor_entity_id)[0]:
            for ref_i in range(len(reference_chains)):
                if reference_entity_ids[ref_i] == anchor_entity_id:
                    yield ref_i, pose_i


def _get_superimposition_transform(
    reference_chain: struc.AtomArray, pose_chain: struc.AtomArray
) -> struc.AffineTransformation:
    """
    Get the transformation (translation and rotation) that superimposes the pose chain
    onto the reference chain.

    Parameters
    ----------
    reference_chain, pose_chain : AtomArray
        The chains to superimpose.

    Returns
    -------
    transform : AffineTransformation
        The transformation that superimposes the pose chain onto the reference chain.
    """
    if is_small_molecule(reference_chain):
        if reference_chain.array_length() == pose_chain.array_length():
            _, transform = struc.superimpose(reference_chain, pose_chain)
        else:
            # The small molecules have different lengths -> difficult superimposition
            # -> simply get an identity transformation
            # This case can only happen in small molecule-only systems anyway
            transform = struc.AffineTransformation(
                center_translation=np.zeros(3),
                rotation=np.eye(3),
                target_translation=np.zeros(3),
            )
    else:
        _, transform, _, _ = struc.superimpose_homologs(
            reference_chain,
            pose_chain,
            _IDENTITY_MATRIX,
            gap_penalty=-1,
            min_anchors=1,
            # No outlier removal
            max_iterations=1,
        )
    return transform


def _is_matched(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    annotation_names: list[str],
) -> bool:
    """
    Check if the given annotations are the same in both structures.

    Parameters
    ----------
    pose, reference : AtomArray
        The pose and reference structure to be compared, respectively.
    annotation_names : list of str
        The names of the annotations to be compared.

    Returns
    -------
    matched : bool
        True, if the annotations are the same in both structures.
    """
    if reference.array_length() != pose.array_length():
        return False
    for annot_name in annotation_names:
        if not (
            reference.get_annotation(annot_name) == pose.get_annotation(annot_name)
        ).all():
            return False
    return True


def _equal_composition(molecules: list[struc.AtomArray]) -> bool:
    """
    Check if the element composition is the same in both molecules.

    The atoms in each molecule may be in a different order.

    Parameters
    ----------
    molecules : list[AtomArray]
        The atoms to compare.

    Returns
    -------
    equal : bool
        True, if the element composition is the same in both molecules.
    """
    element_counters: list[Counter[str]] = [Counter() for _ in molecules]
    for i, molecule in enumerate(molecules):
        for element in molecule.element:
            element_counters[i][element] += 1
    return all(element_counters[0] == counter for counter in element_counters[1:])


def _to_mol(molecule: struc.AtomArray) -> Mol:
    """
    Create a RDKit :class:`Mol` from the given structure and prepare it for usage in
    atom matching.

    Parameters
    ----------
    molecule : struc.AtomArray
        The molecule to convert.

    Returns
    -------
    mol : Mol
        The RDKit molecule.
    """
    mol = rdkit_interface.to_mol(molecule)
    # Make RDKit distinguish stereoisomers when matching atoms
    AssignStereochemistryFrom3D(mol)
    SanitizeMol(
        mol,
        SanitizeFlags.SANITIZE_SETCONJUGATION | SanitizeFlags.SANITIZE_SETAROMATICITY,
    )
    # Make conjugated terminal groups truly symmetric (e.g. carboxyl oxygen atoms)
    for bond in mol.GetBonds():
        if bond.GetIsConjugated() and bond.GetBondType() in [
            BondType.SINGLE,
            BondType.DOUBLE,
        ]:
            bond.SetBondType(BondType.ONEANDAHALF)
    return mol



__all__ = ["Evaluator", "MatchWarning", "EvaluationWarning"]

import copy
import itertools
import warnings
from collections.abc import Iterable, Mapping
from enum import Enum
from typing import Iterator, Sequence
import biotite.structure as struc
import numpy as np
import pandas as pd
from peppr.common import standardize
from peppr.match import find_all_matches, find_optimal_match
from peppr.metric import Metric
from peppr.selector import Selector


class MatchWarning(UserWarning):
    """
    This warning is raised, if a the :class:`Evaluator` fails to match atoms between
    the reference and pose structures.
    """

    pass


class EvaluationWarning(UserWarning):
    """
    This warning is raised, if a :class:`Metric` fails to evaluate a pose.
    """

    pass


class Evaluator(Mapping):
    """
    This class represents the core of :mod:`peppr`.
    Systems are fed via :meth:`feed()` into the :class:`Evaluator`.
    Finally, the evaluation is reported via :meth:`tabulate_metrics()`, which gives a
    scalar metric value for each fed system, or via :meth:`summarize_metrics()`,
    which aggregates the metrics over all systems.

    Parameters
    ----------
    metrics : Iterable of Metric
        The metrics to evaluate the poses against.
        These will make up the columns of the resulting dataframe from
        :meth:`tabulate_metrics()`.
    match_method : MatchMethod, optional
        The strategy to use for finding atom matches between the reference and pose.
        This can be used to trade off speed and accuracy.
    max_matches : int, optional
        The maximum number of atom matches to try, if the `match_method` is set to
        ``EXHAUSTIVE`` or ``INDIVIDUAL``.
    tolerate_exceptions : bool, optional
        If set to true, exceptions during :class:`Metric.evaluate()` are not propagated.
        Instead a warning is raised and the result is set to ``None``.
    min_sequence_identity : float
        The minimum sequence identity for two chains to be considered the same entity.
    allow_unmatched_entities : bool, optional
        If set to ``True``, allow entire entities to be unmatched.
        This is useful if a pose is compared to a reference which may contain different
        molecules.
    remove_monoatomic_ions : bool, optional
        If set to ``True``, monoatomic ions will be removed from the reference and pose
        during :meth:`standardize()`.
    use_entity_annotation : bool, optional
        If set to ``True``, use the ``entity_id`` annotation to determine which chains
        are the same entity and therefore are mappable to each other.
        By default, the entity is determined from sequence identity for polymers and
        residue name for small molecules.
    use_structure_match : bool, optional
        If set to ``True``, use structure matching, i.e. isomorphism on the bond graph,
        to determine which small molecules are the same entity.
        Otherwise, match small molecules by residue name.
        Note that the structure match requires more computation time.

    Attributes
    ----------
    metrics : tuple of Metric
        The metrics to evaluate the poses against.
    system_ids : tuple of str
        The IDs of the systems that were fed into the evaluator.
    """

    class MatchMethod(Enum):
        """
        Method for finding atom matches between the fed reference and pose.
        These methods represent a tradeoff between speed and accuracy.

        - ``HEURISTIC``: Use a fast heuristic [1]_ that matches the reference and pose
          by minimizing the RMSD between the centroids of each chain.
          This method is fast and scales linearly with the number of chains, but
          it is not guaranteed to find the optimal match in all cases, especially
          when the pose and reference are quite distant from each other.
        - ``EXHAUSTIVE``: Exhaustively iterate through all valid atom mappings
          between the reference and pose and select the one that gives the lowest
          all-atom RMSD.
          This method is slower and prone to combinatorial explosion, but it finds
          better matches in edge cases.
        - ``INDIVIDUAL``: Like ``EXHAUSTIVE``, but instead of using the RMSD as
          criterion for optimization, each individual :class:`Metric` is used.
          As this requires exhaustive iteration over all mappings and computing the
          each metric for all of them, this method is slower than ``EXHAUSTIVE``.
          However, it guarantees to find the optimal match for each metric.
        - ``NONE``: Skip atom matching entirely and evaluate metrics on the
          structures as provided. This is useful when the reference and pose are
          already properly aligned, or when using metrics that don't require
          matching (e.g., bond-length violations, clash counts).

        References
        ----------
        .. [1] *Protein complex prediction with AlphaFold-Multimer*, Section 7.3, https://doi.org/10.1101/2021.10.04.463034
        """

        HEURISTIC = "heuristic"
        EXHAUSTIVE = "exhaustive"
        INDIVIDUAL = "individual"
        NONE = "none"

    def __init__(
        self,
        metrics: Iterable[Metric],
        match_method: MatchMethod = MatchMethod.HEURISTIC,
        max_matches: int | None = None,
        tolerate_exceptions: bool = False,
        min_sequence_identity: float = 0.95,
        allow_unmatched_entities: bool = False,
        remove_monoatomic_ions: bool = True,
        use_entity_annotation: bool = False,
        use_structure_match: bool = False,
    ):
        self._metrics = tuple(metrics)
        self._match_method = match_method
        self._max_matches = max_matches
        self._results: list[list[np.ndarray]] = [[] for _ in range(len(metrics))]
        self._ids: list[str] = []
        self._tolerate_exceptions = tolerate_exceptions
        self._min_sequence_identity = min_sequence_identity
        self._allow_unmatched_entities = allow_unmatched_entities
        self._remove_monoatomic_ions = remove_monoatomic_ions
        self._use_entity_annotation = use_entity_annotation
        self._use_structure_match = use_structure_match

    @property
    def metrics(self) -> tuple[Metric, ...]:
        # Use tuple to forbid adding/removing metrics after initialization
        return self._metrics

    @property
    def system_ids(self) -> tuple[str, ...]:
        return tuple(self._ids)

    @staticmethod
    def combine(evaluators: Iterable["Evaluator"]) -> "Evaluator":
        """
        Combine multiple :class:`Evaluator` instances into a single one,
        preserving the systems fed to each instance.

        Parameters
        ----------
        evaluators : Iterable of Evaluator
            The evaluators to combine.
            The ``metrics``, ``tolerate_exceptions`` and ``min_sequence_identity``
            must be the same for all evaluators.

        Returns
        -------
        Evaluator
            The evaluator combining the systems of all input `evaluators` in the order
            of the input.
        """
        ref_evaluator = None
        all_ids = []
        all_results = []
        for evaluator in evaluators:
            if ref_evaluator is None:
                ref_evaluator = evaluator
            if evaluator != ref_evaluator:
                raise ValueError(
                    "All evaluators must be initialized with the same parameters"
                )
            all_ids.append(evaluator.system_ids)
            all_results.append(evaluator.get_results())
        if ref_evaluator is None:
            raise ValueError("At least one evaluator must be provided")

        combined_evaluator = copy.deepcopy(ref_evaluator)
        combined_evaluator._ids = list(itertools.chain(*all_ids))
        combined_evaluator._results = [
            list(itertools.chain(*[results[i] for results in all_results]))
            for i in range(len(combined_evaluator._metrics))
        ]
        return combined_evaluator

    def feed(
        self,
        system_id: str,
        reference: struc.AtomArray,
        poses: Sequence[struc.AtomArray] | struc.AtomArrayStack | struc.AtomArray,
    ) -> None:
        """
        Evaluate the poses of a system against the reference structure for all metrics.

        Parameters
        ----------
        system_id : str
            The ID of the system that was evaluated.
        reference : AtomArray
            The reference structure of the system.
            Each separate instance/molecule must have a distinct `chain_id`.
        poses : AtomArrayStack or list of AtomArray or AtomArray
            The pose(s) to evaluate.
            It is expected that the poses are sorted from highest to lowest confidence,
            (relevant for :class:`Selector` instances).

        Notes
        -----
        `reference` and `poses` must fulfill the following requirements:

        - The system must have an associated `biotite.structure.BondList`,
          i.e. the ``bonds`` attribute must not be ``None``.
        - Each molecule in the system must have a distinct ``chain_id``.
        - Chains where the ``hetero`` annotation is ``True`` is always interpreted as a
          small molecule.
          Conversely, chains where the ``hetero`` annotation is ``False`` is always
          interpreted as protein or nucleic acid chain.
        - Two small molecules can only be matched to each other if they have the same
          ``res_name``.

        The optimal atom matching is handled automatically based on the
        :class:`MatchMethod`.
        """
        try:
            reference = standardize(
                reference, remove_monoatomic_ions=self._remove_monoatomic_ions
            )
            if isinstance(poses, struc.AtomArray):
                poses = [
                    standardize(
                        poses, remove_monoatomic_ions=self._remove_monoatomic_ions
                    )
                ]
            elif isinstance(poses, struc.AtomArrayStack):
                poses = list(
                    standardize(
                        poses, remove_monoatomic_ions=self._remove_monoatomic_ions
                    )
                )
            else:
                poses = [
                    standardize(
                        pose, remove_monoatomic_ions=self._remove_monoatomic_ions
                    )
                    for pose in poses
                ]
            if len(poses) == 0:
                raise ValueError("No poses provided")
        except Exception as e:
            self._raise_or_warn(
                e,
                UserWarning(f"Failed to standardize system '{system_id}': {e}"),
            )

        if self._match_method in (
            Evaluator.MatchMethod.HEURISTIC,
            Evaluator.MatchMethod.EXHAUSTIVE,
        ):
            use_heuristic = self._match_method == Evaluator.MatchMethod.HEURISTIC
            result_for_system = np.stack(
                [
                    self._evaluate_using_rmsd(system_id, reference, pose, use_heuristic)
                    for pose in poses
                ],
                axis=-1,
            )
        elif self._match_method == Evaluator.MatchMethod.INDIVIDUAL:
            result_for_system = np.stack(
                [
                    self._evaluate_using_each_metric(system_id, reference, pose)
                    for pose in poses
                ],
                axis=-1,
            )
        elif self._match_method == Evaluator.MatchMethod.NONE:
            result_for_system = np.stack(
                [
                    self._evaluate_without_matching(system_id, reference, pose)
                    for pose in poses
                ],
                axis=-1,
            )
        else:
            raise ValueError(f"Unknown match strategy: {self._match_method}")

        for i, result in enumerate(result_for_system):
            self._results[i].append(result)
        self._ids.append(system_id)

    def get_results(self) -> list[list[np.ndarray]]:
        """
        Return the raw results of the evaluation.

        This includes each metric evaluated on each pose of each system.

        Returns
        -------
        list of list of np.ndarray
            The raw results of the evaluation.
            The outer list iterates over the metrics, the inner list iterates over
            the systems and the array represents the values for each pose.
        """
        return copy.deepcopy(self._results)

    def tabulate_metrics(
        self, selectors: Iterable[Selector] | None = None
    ) -> pd.DataFrame:
        """
        Create a table listing the value for each metric and system.

        Parameters
        ----------
        selectors : list of Selector, optional
            The selectors to use for selecting the best pose of a multi-pose
            prediction.
            This parameter is not necessary if only single-pose predictions were fed
            into the :class:`Evaluator`.

        Returns
        -------
        pandas.DataFrame
            A table listing the value for each metric and system.
            The index is the system ID.
        """
        columns = self._tabulate_metrics(selectors)
        # Convert (metric, selector)-tuples to strings
        columns = {
            (
                f"{metric.name} ({selector.name})"
                if selector is not None
                else metric.name
            ): values
            for (metric, selector), values in columns.items()
        }
        return pd.DataFrame(columns, index=self._ids)

    def summarize_metrics(
        self, selectors: Iterable[Selector] | None = None
    ) -> dict[str, float]:
        """
        Condense the system-wise evaluation to scalar values for each metric.

        For each metric,

        - the mean value
        - the median value
        - and the percentage of systems within each threshold

        is computed.

        Parameters
        ----------
        selectors : list of Selector, optional
            The selectors to use for selecting the best pose of a multi-pose
            prediction.
            This parameter is not necessary if only single-pose predictions were fed
            into the :class:`Evaluator`.

        Returns
        -------
        dict (str -> float)
            A dictionary mapping the summarized metric name to the scalar value.
            The summarized metric name contains

            - the metric name (e.g. ``DockQ``)
            - the selector name, if a selector was used (e.g. ``Oracle``)
            - the threshold (if a threshold was used) (e.g. ``% acceptable``)
        """
        columns = self._tabulate_metrics(selectors)
        output_columns = {}
        for (metric, selector), values in columns.items():
            if metric.thresholds:
                edges = list(metric.thresholds.values()) + [np.inf]
                counts_per_bin, _ = np.histogram(values, bins=edges)
                # NaN values do not bias the percentages,
                # as they are not included in any bin
                percentages_per_bin = counts_per_bin / np.sum(counts_per_bin)
                for threshold_name, percentage in zip(
                    metric.thresholds.keys(), percentages_per_bin
                ):
                    column_name = f"{metric.name} {threshold_name}"
                    if selector is not None:
                        column_name += f" ({selector.name})"
                    output_columns[column_name] = percentage.item()
            # Always add the mean and median value as well
            for name, function in [("mean", np.nanmean), ("median", np.nanmedian)]:
                column_name = f"{metric.name} {name}"
                if selector is not None:
                    column_name += f" ({selector.name})"
                output_columns[column_name] = function(values).item()  # type: ignore[operator]
        return output_columns

    def __getitem__(self, metric_name: str) -> list[np.ndarray]:
        return self._results[self._metrics.index(metric_name)]

    def __iter__(self) -> Iterator[list[np.ndarray]]:
        return iter(self._results)

    def __len__(self) -> int:
        return len(self._results)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Evaluator):
            return False
        if len(self._metrics) != len(other._metrics):
            return False
        for metric, ref_metric in zip(self._metrics, other._metrics):
            if metric.name != ref_metric.name:
                return False
        if self._tolerate_exceptions != other._tolerate_exceptions:
            return False
        if self._min_sequence_identity != other._min_sequence_identity:
            return False
        return True

    def _evaluate_using_rmsd(
        self,
        system_id: str,
        reference: struc.AtomArray,
        pose: struc.AtomArray,
        use_heuristic: bool,
    ) -> np.ndarray:
        """
        Run the metrics on the given system using matching that aims to minimize the
        RMSD between the reference and pose.
        """
        results = np.full(len(self._metrics), np.nan)
        try:
            matched_reference, matched_pose = find_optimal_match(
                reference,
                pose,
                min_sequence_identity=self._min_sequence_identity,
                use_heuristic=use_heuristic,
                max_matches=self._max_matches,
                allow_unmatched_entities=self._allow_unmatched_entities,
                use_entity_annotation=self._use_entity_annotation,
                use_structure_match=self._use_structure_match,
            )
        except Exception as e:
            self._raise_or_warn(
                e,
                MatchWarning(
                    f"Failed to match reference and pose in system '{system_id}': {e}"
                ),
            )
            return results

        for i, metric in enumerate(self._metrics):
            try:
                results[i] = metric.evaluate(matched_reference, matched_pose)
            except Exception as e:
                self._raise_or_warn(
                    e,
                    EvaluationWarning(
                        f"Failed to evaluate {metric.name} on '{system_id}': {e}"
                    ),
                )
        return results

    def _evaluate_without_matching(
        self, system_id: str, reference: struc.AtomArray, pose: struc.AtomArray
    ) -> np.ndarray:
        """
        Run the metrics on the given system without performing any atom matching.

        This assumes that the reference and pose structures are already properly
        aligned with matching atom ordering.
        """
        results = np.full(len(self._metrics), np.nan)

        for i, metric in enumerate(self._metrics):
            try:
                results[i] = metric.evaluate(reference, pose)
            except Exception as e:
                self._raise_or_warn(
                    e,
                    EvaluationWarning(
                        f"Failed to evaluate {metric.name} on '{system_id}': {e}"
                    ),
                )
        return results

    def _evaluate_using_each_metric(
        self, system_id: str, reference: struc.AtomArray, pose: struc.AtomArray
    ) -> np.ndarray:
        """
        Run the metrics on the given system using a separate matching for each metric,
        that specifically optimizes the given metric.
        """

        class _FindAllMatchesError(Exception):
            """
            An exception class to be able to specifically catch exceptions in the
            loop head calling `find_all_matches()`.

            Parameters
            ----------
            wrapped_exception : Exception
                The actual exception raised by `find_all_matches()`.
            """

            def __init__(self, wrapped_exception):  # type: ignore[no-untyped-def]
                self.wrapped_exception = wrapped_exception

        def _find_all_matches(*args, **kwargs):  # type: ignore[no-untyped-def]
            try:
                for e in find_all_matches(*args, **kwargs):
                    yield e
            except Exception as e:
                raise _FindAllMatchesError(e)

        results = np.full(len(self._metrics), np.nan)

        for i, metric in enumerate(self._metrics):
            try:
                best_result = np.inf if metric.smaller_is_better() else -np.inf
                for it, (matched_reference, matched_pose) in enumerate(
                    _find_all_matches(
                        reference,
                        pose,
                        min_sequence_identity=self._min_sequence_identity,
                        allow_unmatched_entities=self._allow_unmatched_entities,
                        use_entity_annotation=self._use_entity_annotation,
                        use_structure_match=self._use_structure_match,
                    )
                ):
                    if self._max_matches is not None and it >= self._max_matches:
                        break
                    try:
                        result = metric.evaluate(matched_reference, matched_pose)
                    except Exception as e:
                        self._raise_or_warn(
                            e,
                            EvaluationWarning(
                                f"Failed to evaluate {metric.name} "
                                f"on '{system_id}': {e}"
                            ),
                        )
                        best_result = np.nan
                        break
                    if metric.smaller_is_better():
                        if result < best_result:
                            best_result = result
                    else:
                        if result > best_result:
                            best_result = result

            except _FindAllMatchesError as e:
                self._raise_or_warn(
                    e.wrapped_exception,
                    MatchWarning(
                        f"Failed to match reference and pose "
                        f"in system '{system_id}': {e}"
                    ),
                )
                best_result = np.nan

            if np.isfinite(best_result):
                results[i] = best_result

        return results

    def _tabulate_metrics(
        self, selectors: Iterable[Selector] | None = None
    ) -> dict[tuple[Metric, Selector | None], np.ndarray]:
        columns: dict[tuple[Metric, Selector | None], np.ndarray] = {}
        for i, metric in enumerate(self._metrics):
            values = self._results[i]
            if not selectors:
                condensed_values = []
                for array in values:
                    if array is None:
                        condensed_values.append(np.nan)
                    elif len(array) > 1:
                        raise ValueError(
                            "At least one selector is required for multi-pose predictions"
                        )
                    else:
                        condensed_values.append(array[0])
                columns[metric, None] = np.array(condensed_values)
            else:
                for selector in selectors:
                    condensed_values = np.array(
                        [
                            selector.select(val, metric.smaller_is_better())
                            if val is not None
                            else np.nan
                            for val in values
                        ]
                    )  # type: ignore[assignment]
                    columns[metric, selector] = condensed_values  # type: ignore[assignment]
        return columns

    def _raise_or_warn(
        self, exception: Exception, alternative_warning: Warning | type[Warning]
    ) -> None:
        """
        Raise the given exception, if ``tolerate_exceptions`` is set to ``False``,
        or raise a warning otherwise.

        Parameters
        ----------
        exception : Exception
            The exception to raise.
        alternative_warning : Warning or type[Warning]
            The warning to raise if ``tolerate_exceptions`` is set to ``False``.
            If only :class:`Warning` type is given instead of an instance, the warning
            message is taken from the `exception`.
        """
        if self._tolerate_exceptions:
            if isinstance(alternative_warning, type):
                warnings.warn(str(exception), alternative_warning)
            else:
                warnings.warn(alternative_warning)
        else:
            raise exception



__all__ = ["estimate_formal_charges"]

import functools
from pathlib import Path
import biotite.interface.rdkit as rdkit_interface
import biotite.structure as struc
import numpy as np
import pandas as pd
import rdkit.Chem.AllChem as Chem
from numpy.typing import NDArray
from peppr.sanitize import sanitize

METAL_CHARGES = {
    # For metals with multiple oxidation states, the smallest charge is assumed
    # For the computation of salt bridges it is only important that an atom has a
    # positive charge, anyway
    "LI": 1,
    "K": 1,
    "NA": 1,
    "RB": 1,
    "BE": 2,
    "MG": 2,
    "CA": 2,
    "CR": 2,
    "MO": 3,
    "MN": 2,
    "FE": 2,
    "RU": 3,
    "CO": 2,
    "NI": 2,
    "CU": 1,
    "ZN": 2,
    "AL": 3,
}


def estimate_formal_charges(
    atoms: struc.AtomArray, ph: float = 7.4
) -> NDArray[np.int_]:
    """
    Determine the formal charge of each atom in the structure.

    Parameters
    ----------
    atoms : AtomArray
        The atoms to determine the formal charge for.
        Must have an associated :class:`BondList`.
    ph : float, optional
        The pH of the environment.
        By default a physiological pH value is used [1]_.

    Returns
    -------
    np.ndarray, shape=(n,)
        The formal charge of each atom.

    References
    ----------
    .. [1] https://www.ncbi.nlm.nih.gov/books/NBK507807/
    """

    patterns = []
    for row in _get_protonation_table().itertuples():
        if _is_active(ph, row.ph):
            patterns.append((_get_pattern(row.group), float(row.charge)))
        else:
            # neutralize when inactive!
            patterns.append((_get_pattern(row.group), 0.0))

    mol = rdkit_interface.to_mol(atoms)
    sanitize(mol)
    # Initialize with the original charges
    charges = rdkit_interface.from_mol(mol, add_hydrogen=False).charge
    # Remove hydrogen atoms to correctly match SMARTS atoms with the '[D<n>]' property
    mol = Chem.RemoveAllHs(mol, sanitize=False)  # type: ignore[attr-defined]

    for pattern, charge in patterns:
        matches = mol.GetSubstructMatches(pattern)
        for match in matches:
            if len(match) == 1:
                # There is only one matched atom -> assign charge to it
                atom_index = match[0]
            elif (index := _get_atom_index_for_map(pattern, 1)) is not None:
                # Alternatively, use the atom with map number 1 ('[X:1]' in SMARTS)
                atom_index = match[index]
            else:
                raise ValueError(
                    "The pattern does not specify the atom to assign the charge to"
                )
            charges[atom_index] = charge

    # Assign charges to metals
    metal_charges = np.array(
        [METAL_CHARGES.get(element.upper(), 0) for element in atoms.element]
    )
    charges[metal_charges != 0] = metal_charges[metal_charges != 0]

    return charges


@functools.cache
def _get_protonation_table() -> pd.DataFrame:
    """
    Load the table that determines the protonation/deprotonation reactions based on pH.

    Returns
    -------
    table : pd.DataFrame
        The table.
    """
    return pd.read_csv(
        Path(__file__).parent / "protonation.csv",
        delimiter=";",
        dtype=str,
        index_col=False,
        comment="#",
        keep_default_na=False,  # Use literally '' for empty pH condition
    )


@functools.cache
def _get_pattern(group_pattern: str) -> Chem.Mol:
    """
    Create the :class:`Mol` from SMARTS pattern.

    Parameters
    ----------
    group_pattern : str
        The SMARTS pattern.

    Returns
    -------
    mol : Chem.Mol
        The pattern as :class:`Mol`.
    """
    return Chem.MolFromSmarts(group_pattern)  # type: ignore[attr-defined]


def _is_active(ph: float, ph_condition: str) -> bool:
    """
    Check if the protonation/deprotonation should occur at the given pH.

    Parameters
    ----------
    ph : float
        The pH.
    ph_condition : str
        The pH condition, e.g. ``<7.4``.

    Returns
    -------
    reacts : bool
        If the protonation/deprotonation should occur.
        If the pH condition is empty, always return ``True``.
    """
    ph_condition = ph_condition.strip().replace("=", "")
    if ph_condition == "":
        return True
    elif ph_condition.startswith("<"):
        return ph < float(ph_condition[1:])
    elif ph_condition.startswith(">"):
        return ph > float(ph_condition[1:])
    else:
        raise ValueError(f"Invalid pH condition: {ph_condition}")


def _get_atom_index_for_map(pattern: Chem.Mol, map_number: int) -> int | None:
    """
    Get the index of the atom with the given map number.

    Parameters
    ----------
    pattern : Chem.Mol
        The pattern.
    map_number : int
        The map number.

    Returns
    -------
    atom_index : int or None
        The index of the atom with the given map number.
        If no atom with the given map number exists, return ``None``.
    """
    if map_number < 1:
        raise ValueError("Map number must be greater than 0")
    for atom in pattern.GetAtoms():
        map_num = atom.GetAtomMapNum()
        if map_num == map_number:
            return atom.GetIdx()
    return None



__all__ = ["volume", "volume_overlap"]

import biotite.structure as struc
import biotite.structure.info as info
import numpy as np
from numpy.typing import NDArray


def volume(molecule: struc.AtomArray, voxel_size: float = 0.5) -> float:
    """
    Calculate the volume of the given molecule.

    Parameters
    ----------
    molecule : struc.AtomArray
        The molecule to calculate the volume of.
    voxel_size : float, optional
        The size of the voxels used for volume calculation.
        The computation becomes more accurate with smaller voxel sizes, but
        the run time scales inverse cubically with voxel size.

    Returns
    -------
    float
        The volume of the molecule.
    """
    voxel_volume = voxel_size**3
    vdw_radii = np.array(
        [info.vdw_radius_single(element) for element in molecule.element]
    )
    max_vdw_radius = np.max(vdw_radii)
    voxel_grid = _create_voxel_grid(molecule, voxel_size, max_vdw_radius)
    cell_list = struc.CellList(voxel_grid, max_vdw_radius)
    occupied_mask = _find_occupied_voxels(cell_list, voxel_grid, molecule, vdw_radii)
    return voxel_volume * np.count_nonzero(occupied_mask)


def volume_overlap(
    molecules: list[struc.AtomArray],
    voxel_size: float = 0.5,
) -> tuple[NDArray[np.floating], float, float]:
    """
    Calculate the volume of the given molecules and how their volumes overlap
    (i.e. their intersection and union).

    Parameters
    ----------
    molecules : list of struc.AtomArray, length=n
        The molecules to calculate the volume of.
    voxel_size : float, optional
        The size of the voxels used for volume calculation.
        The computation becomes more accurate with smaller voxel sizes, but
        the run time scales inverse cubically with voxel size.

    Returns
    -------
    volumes : np.ndarray, shape=(n,), dtype=float
        The volume of each input molecule.
    intersection_volume : float
        The volume intersection of all molecules.
    union_volume : float
        The volume union of all molecules.
    """
    voxel_volume = voxel_size**3

    vdw_radii = [
        np.array([info.vdw_radius_single(element) for element in molecule.element])
        for molecule in molecules
    ]
    max_vdw_radius = max(np.max(radii) for radii in vdw_radii)

    voxel_grids = [
        _create_voxel_grid(molecule, voxel_size, max_vdw_radius)
        for molecule in molecules
    ]
    common_voxel_grid = np.unique(np.concatenate(voxel_grids, axis=0), axis=0)
    cell_list = struc.CellList(common_voxel_grid, max_vdw_radius)
    occupied_masks = [
        _find_occupied_voxels(cell_list, common_voxel_grid, molecule, radii)
        for molecule, radii in zip(molecules, vdw_radii)
    ]
    volumes = [
        voxel_volume * np.count_nonzero(occupied_mask)
        for occupied_mask in occupied_masks
    ]
    intersection_volume = voxel_volume * np.count_nonzero(
        np.all(occupied_masks, axis=0)
    )
    union_volume = voxel_volume * np.count_nonzero(np.any(occupied_masks, axis=0))

    return np.array(volumes), intersection_volume, union_volume


def _create_voxel_grid(
    molecule: struc.AtomArray, voxel_size: float, padding: float
) -> NDArray[np.floating]:
    """
    Create a voxel grid for the given molecule.

    Parameters
    ----------
    molecule : struc.AtomArray
        The molecule to create the voxel grid around.
    voxel_size : float
        The size of the voxels.
    padding : float
        The 'buffer' around the molecule.
        This makes sure that voxels span not only the molecule centers, but also
        their radii.

    Returns
    -------
    voxel_grid : np.ndarray, shape=(n, 3)
        The voxel coordinates.
    """
    min_coord = np.min(molecule.coord, axis=0) - padding
    # Round to the nearest multiple of the voxel size
    # so multiple grids can be combined later using `np.unique()`
    min_voxel_coord = min_coord - min_coord % voxel_size
    max_coord = np.max(molecule.coord, axis=0) + padding
    return np.stack(
        [
            ordinate.ravel()
            for ordinate in np.meshgrid(
                *[
                    np.arange(min_c, max_c, voxel_size)
                    for min_c, max_c in zip(min_voxel_coord, max_coord)
                ]
            )
        ],
        axis=-1,
    )


def _find_occupied_voxels(
    cell_list: struc.CellList,
    voxel_grid: NDArray[np.floating],
    molecule: struc.AtomArray,
    vdw_radii: NDArray[np.floating],
) -> NDArray[np.bool_]:
    """
    Find the voxels that are occupied by the given molecule.

    Parameters
    ----------
    cell_list : struc.CellList
        The cell list that contains the voxel grid.
    voxel_grid : np.ndarray, shape=(n, 3)
        The voxel grid.
    molecule : struc.AtomArray, shape=(k,)
        The molecule to find the occupied voxels for.
    vdw_radii : np.ndarray, shape=(k,), dtype=float
        The van der Waals radii of each atom.

    Returns
    -------
    occupied_voxels : np.ndarray, shape=(n,), dtype=bool
        As mask that is True for each occupied voxel in `voxel_grid`.
    """
    # These voxels are potentially occupied by the molecule...
    atom_indices, voxel_indices = _to_sparse_indices(
        cell_list.get_atoms_in_cells(molecule.coord)
    ).T
    # ... but we need to check if they are within the VdW radius to be sure
    # Use the squared distance to avoid costly square root calculation
    occupied_pair_mask = (
        _squared_distance(molecule.coord[atom_indices], voxel_grid[voxel_indices])
        < vdw_radii[atom_indices] ** 2
    )
    occupied_voxel_indices = voxel_indices[occupied_pair_mask]
    occupied_voxel_mask = np.zeros(voxel_grid.shape[0], dtype=bool)
    occupied_voxel_mask[occupied_voxel_indices] = True
    return occupied_voxel_mask


def _to_sparse_indices(all_contacts: NDArray[np.int_]) -> NDArray[np.int_]:
    """
    Create tuples indices from the :meth:`CellList.get_atoms_in_cells()` return value.

    In other words, they would mark the non-zero elements in a dense contact matrix.

    Parameters
    ----------
    all_contacts : ndarray, dtype=int, shape=(m,n)
        The contact indices as returned by :meth:`CellList.get_atoms_in_cells()`.
        Padded with -1, in the second dimension.
        Dimension *m* marks the query, dimension *n* marks the voxels in the cell list.

    Returns
    -------
    combined_indices : ndarray, dtype=int, shape=(l,2)
        The contact indices.
        Each column contains the query and contact voxel index.
    """
    # Find rows where a query atom has at least one contact
    non_empty_indices = np.where(np.any(all_contacts != -1, axis=1))[0]
    # Take those rows and flatten them
    contact_indices = all_contacts[non_empty_indices].flatten()
    # For each row the corresponding query atom is the same
    # Hence in the flattened form the query atom index is simply repeated
    query_indices = np.repeat(non_empty_indices, all_contacts.shape[1])
    combined_indices = np.stack([query_indices, contact_indices], axis=1)
    # Remove the padding values
    return combined_indices[contact_indices != -1]


def _squared_distance(
    a: NDArray[np.floating], b: NDArray[np.floating]
) -> NDArray[np.floating]:
    """
    Compute the squared distance between two points.

    Parameters
    ----------
    a, b : np.ndarray, shape=(n, 3)
        The points.

    Returns
    -------
    distance : float
        The squared distance between the two points.
    """
    return np.sum((a - b) ** 2, axis=1)
